from __future__ import annotations

import sys
from itertools import combinations_with_replacement
from typing import Callable

import numpy as np
import torch

if sys.version_info >= (3, 8):
    pass
else:
    pass


class BaseRegressionModel(torch.nn.Module):
    """
    Base model for regression.
    
    :ivar w: weights of model
    :ivar _best_state: _best_state while model training
    :ivar _best_loss: _best_loss while model training
    """

    def __init__(self) -> None:
        """
        Initialization of base model for different regression
        """
        super(BaseRegressionModel, self).__init__()
        self.w = None
        self._best_state = None
        self._best_loss = torch.inf
        self.bias = True

    @staticmethod
    def transform(x: torch.Tensor) -> torch.Tensor:
        """
        Returns transformed x. Default is return x without changes

        :param x: torch tensor
        :return: transformed torch tensor
        """
        return x

    def forward(self, x: torch.Tensor, transformed: bool = True) -> torch.Tensor:
        """
        Returns transform(x) @ w, w is the weights for each parameter, transform(x) is some transformed matrix.

        #. Linear transformed - same matrix
        #. Polynomial transform check polynomial
        #. Exponential transform check exponential

        For linear returns x @ w + b, w is the weights for each parameter, and b is the bias

        .. math::

            \\hat Y_{n \\times 1} = X_{n \\times m} \\cdot W_{m \\times 1} + b \\cdot I_{n \\times 1} =
            \\begin{bmatrix}
            w_1 x_{1, 1} + w_2 x_{1, 2} + \\dots + w_m + x_{1, m} + b\\\\
            \\vdots \\\\
            w_1 x_{n, 1} + w_2 x_{n, 2} + \\dots + w_m + x_{n, m} + b \\\\
            \\end{bmatrix}

        For non linear:

        .. math::

            \\hat Y_{n \\times 1} = X_{\\operatorname{transformed}} \\cdot W

        :param x: input observations, tensor n x m (n is the number of observations that have m parameters)
        :param transformed: flag of transformed x
        :return: regression value
        """

        if self.w is None:
            self.init_weights(x)

        if not transformed:
            x = self.transform(x)

        return self.w(x)

    _forward = forward

    def init_weights(self, x: torch.Tensor) -> None:
        """
        Initializing weights

        :param x: input observations, tensor n x m (n is the number of observations that have m parameters)
        """

        self.w = torch.nn.Linear(x.shape[1], 1, bias=self.bias)
        self._best_state = self.w.state_dict()

    def fit(self,
            x: torch.Tensor,
            y: torch.Tensor,
            epochs: int = 2000,
            lr: float = 1e-4,
            l1_constant: float = 0.,
            l2_constant: float = 0.,
            show_epoch: int = 0,
            print_function: Callable = print) -> torch.nn.Module:
        """
        Returns trained model of Regression

        .. rubric:: Target function

        Training happens by minimizing loss function:

        .. math::

            \\mathcal{L}(w) = \\lambda_{1} \\Vert w \\Vert_{1} + \\lambda_{2} \\Vert w \\Vert_{2} +
            \\frac{1}{n}\\sum_{i = 1}^{n} (\\hat y_i - y_i)^2 \\longrightarrow \\min_{w}

        :math:`x_{i} \\in \\mathbb{R}^{1\\times m}, w \\in \\mathbb{R}^{m \\times 1}, y_{i} \\in \\mathbb{R}^{1}`

        :param x: training set
        :param y: target value
        :param epochs: max number of sgd implements
        :param lr: Adam optimizer learning rate
        :param show_epoch: amount of showing epochs
        :param l1_constant: parameter of l1 regularization
        :param l2_constant: parameter of l2 regularization
        :param print_function: a function that will print verbose
        :return: trained model
        """
        x = x.float()
        x = self.transform(x)
        y = y.float().flatten()

        return self._training(x, y, epochs, lr, l1_constant, l2_constant, show_epoch, print_function)

    def _training(self,
                  x: torch.Tensor,
                  y: torch.Tensor,
                  epochs: int = 2000,
                  lr: float = 1e-4,
                  l1_constant: float = 0.,
                  l2_constant: float = 0.,
                  show_epoch: int = 0,
                  print_function: Callable = print) -> torch.nn.Module:
        """
        Gradient descent to minimize loss function

        :param x: training set
        :param y: target value
        :param epochs: max number of sgd implements
        :param lr: Adam optimizer learning rate
        :param show_epoch: amount of showing epochs
        :param l1_constant: parameter of l1 regularization
        :param l2_constant: parameter of l2 regularization
        :param print_function: a function that will print verbose
        :return:
        """
        self.forward(x)  # model initializing

        # set starting point from by analytical formula
        x_t = PolyTransform(1).fit_transform(x)

        try:
            b = torch.linalg.inv(x_t.T @ x_t) @ x_t.T @ y
            _w = list(self.w.parameters())
            _w[0].data = b[1:].reshape(1, -1)
            _w[1].data = b[0].reshape(1, -1)

        except torch.linalg.LinAlgError:
            pass

        criterion = torch.nn.MSELoss()
        optimizer = torch.optim.Adam(self.parameters(), lr=lr, weight_decay=l2_constant)
        print_epochs = np.linspace(1, epochs, min(epochs, show_epoch), dtype=int)

        for epoch in range(1, epochs + 1):
            optimizer.zero_grad()
            output = criterion(self._forward(x).flatten(), y.flatten())
            if output.item() < self._best_loss:
                self._best_state = self.state_dict()

            if l1_constant > 0:
                for layer in self.parameters():
                    output += l1_constant * layer.data.sum()

            output.backward()
            optimizer.step()

            with torch.no_grad():
                if epoch + 1 in print_epochs:
                    print_function(f'Epoch: {epoch: 5d} | MSE: {output.item(): 0.5f}')

            if output.item() < 1e-10:
                if show_epoch > 0:
                    print_function(f'Epoch: {epoch: 5d} | MSE < 1e-10 | Early stop')
        else:
            self.load_state_dict(self._best_state)

        return self


class LinearRegression(BaseRegressionModel):
    """
    Model:

    .. math::

        \\hat y(x) = w_0 + w_1 \\cdot x_1 + w_2 \\cdot x_2 + \\dots + w_m \\cdot x_m

    """

    def __init__(self, bias: bool = True) -> None:
        super(LinearRegression, self).__init__()
        self.bias = bias


class PolynomialRegression(BaseRegressionModel):
    """

    Polynomial regression model:

    .. math::
        \\hat y(x) = \\sum_{\\alpha_1 + \\dots + \\alpha_m \\leq k} w_i \\cdot x_1 ^ {\\alpha_1} \\circ
        x_2 ^ {\\alpha_2} \\cdot \\dots \\circ x_m ^ {\\alpha_m}

    :math:`\\alpha_i \\in \\mathbb{Z}_+, x_i - i` column from :math:`x` matrix

    :math:`x_i, y, \\hat y \\in \\mathbb{R}^{n \\times 1}, \\circ` - hadamard product (like np.array * np.array)

    """

    def __init__(self, degree: int) -> None:
        """
        :param degree: degree of polynomial regression
        """
        super().__init__()
        assert degree > 0, 'degree must be positive'
        self.degree = int(degree)
        self.transformer = PolyTransform(self.degree)

    def transform(self, x: torch.Tensor) -> torch.Tensor:
        """
        Returns poly-transformed data

        :param x: torch tensor
        :return: transformed tensor
        """
        if not self.transformer.is_fitted:
            self.transformer.fit(x)

        return self.transformer.transform(x)[:, 1:]

    def __call__(self, x: torch.Tensor, transformed: bool = False) -> torch.Tensor:
        """
        Returns x_pf @ w , w is the weights for each parameter, x_pf is poly transformed matrix

        :param x: input observations, tensor n x m (n is the number of observations that have m parameters)
        :param transformed: flag of transformed x
        :return: regression value
        """
        return self.forward(x, transformed)


class ExponentialRegression(BaseRegressionModel):
    """

    Exponential regression

    .. math::

        \\hat y_i = \\exp(w_0 + w_1 \\cdot x_1 + w_2 \\cdot x_2 + \\dots + w_m \\cdot x_m)

    """

    def __init__(self):
        super(ExponentialRegression, self).__init__()

    def forward(self, x: torch.Tensor, transformed: bool = True) -> torch.Tensor:
        """
        Returns exponential regression function

        .. math::
            \\hat y = \\exp (x \\cdot w + b)

        :param x: input observations, tensor n x m (n is the number of observations that have m parameters)
        :param transformed: flag of transformed x
        :return: regression value
        """
        if self.w is None:
            self.init_weights(x)

        if not transformed:
            x = self.transform(x)

        return torch.exp(self.w(x))

    def fit(self,
            x: torch.Tensor,
            y: torch.Tensor,
            epochs: int = 5000,
            lr: float = 1e-4,
            l1_constant: float = 0.,
            l2_constant: float = 0.,
            show_epoch: int = 0,
            print_function: Callable = print) -> torch.nn.Module:
        """
        Returns trained model of exponential Regression

        :param x: training set
        :param y: target value
        :param epochs: max number of sgd implements
        :param lr: Adam optimizer learning rate
        :param show_epoch: amount of showing epochs
        :param l1_constant: parameter of l1 regularization
        :param l2_constant: parameter of l2 regularization
        :param print_function: a function that will print verbose
        :return: trained model
        """
        x = x.float()
        assert (y > 0).all(), 'y must be positive'
        y = torch.log(y.float().flatten())

        self._forward(x)  # model initializing

        return self._training(x, y, epochs, lr, l1_constant, l2_constant, show_epoch)


class PolyTransform:
    """
    Makes polynomial features of degree :math:`k`

    We have matrix :math:`x \\in \\mathbb{R}^{n \\times m}` with m features and n observations.
    Let :math:`c_1` -- first column of :math:`x`, :math:`c_2` -- second column of :math:`x` etc.
    And we make all features like:

    .. math::

        c_1^{\\alpha_1} \\circ c_2^{\\alpha_2} \\circ \\dots \\circ c_m^{\\alpha_m}

    subject to :math:`\\alpha_i \\in \\mathbb{Z}_+` and :math:`\\sum_{i = 1}^{m} \\alpha_i \\leq k`

    :ivar feature_names: a list with the names of objects and degrees in order, as in the matrix
    :ivar columns: indexes of columns after fitting
    :ivar number_features: number of features after transform

    """

    def __init__(self, degree: int = 2) -> None:
        """
        :param degree: max degree of polynomial features
        """
        self.degree = degree
        self.feature_names = []
        self.columns = []
        self.number_features = None
        self.is_fitted = False

    def fit(self, x: torch.Tensor, columns: list[str] | None = None) -> PolyTransform:
        """
        Fitting of transformer. After fitting appears feature_names, columns, number_features

        :param x: matrix with m columns
        :param columns: string names of each column
        :return: self instance
        """

        column_indexes = list(range(x.shape[1]))

        # make feature names from column names
        if columns is not None:
            for k in range(self.degree + 1):
                self.feature_names.extend(combinations_with_replacement(columns, k))

        # make indexes sets
        for k in range(self.degree + 1):
            self.columns.extend(combinations_with_replacement(column_indexes, k))

        # make columns from indexes
        if columns is None:
            for names in self.columns:
                self.feature_names.append(tuple(map(lambda _x: 'col_' + str(_x), names)))

        feat = ['1']

        # make feature names with
        for i in range(1, len(self.feature_names)):
            val, cnt = np.unique(self.feature_names[i], return_counts=True)
            val = list(map(str, val))
            cnt = list(map(str, cnt))

            feat.append(' '.join(map(lambda _x: '^'.join(_x), zip(val, cnt))))

        self.feature_names = feat
        self.is_fitted = True
        self.number_features = len(self.feature_names)

        return self

    def transform(self, x: torch.Tensor) -> torch.Tensor:
        """
        Returns transformed input x by the indexes from self columns list

        :param x: Tensor with m columns like in fitting
        :return: transformed tensor
        """
        transformed = [torch.ones(x.shape[0], 1)]

        for cols in self.columns[1:]:
            transformed.append(x[:, cols].prod(dim=1, keepdim=True))

        return torch.concat(transformed, dim=1)

    def fit_transform(self, x: torch.tensor) -> torch.Tensor:
        """
        Returns transformed x. Call fit and after transform

        :param x: Tensor with m columns like in fitting
        :return: transformed tensor
        """
        self.fit(x)
        return self.transform(x)


if __name__ == '__main__':
    from metrics import r2_score

    torch.random.manual_seed(7)
    model = LinearRegression()
    _x = torch.rand(1000, 4)
    _y = _x @ torch.randint(0, 10, (4, 1)).float() + 5

    model.fit(_x, _y)

    print(r2_score(model(_x).flatten(), _y.flatten()))

    model2 = PolynomialRegression(3)
    model2.fit(_x, _y)

    print(r2_score(model2(_x).flatten(), _y.flatten()))

    model3 = ExponentialRegression()
    _y = torch.exp(_y)

    model3.fit(_x, _y)
    print(r2_score(model3(_x).flatten(), _y.flatten()))



from __future__ import annotations

from typing import Callable

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import torch

from ..one_optimize import golden_section_search, successive_parabolic_interpolation, brent
from ..support import HistoryGSS, HistorySPI, HistoryBrent

# Color palette
# https://coolors.co/94affd-ff616b-0a50ff-7f8082-50c878

COLOR1 = FRENCH_SKY_BLUE = '#94AFFD'
COLOR2 = FIERY_ROSE = '#FF616B'
COLOR3 = BLUE_RYB = '#0a50ff'
COLOR4 = GRAY_WEB = '#7F8082'
COLOR5 = EMERALD = '#50C878'

standard_layout = dict(
    xaxis_title=r'<b>x</b>',
    yaxis_title=r'<b>f(x)</b>',
    font=dict(size=14),
    legend_title_text='<b>variable names</b>'
)


def transfer_history_gss(history: HistoryGSS, func) -> pd.DataFrame:
    """
    Generate data for plotly express with using animation_frame for animate

    .. code-block:: python3

        >>> def f(x): return x ** 2
        >>> _, hist = golden_section_search(f, (-1, 2), keep_history=True)
        >>> data_for_plot = transfer_history_gss(hist, f)
        Searching finished. Successfully. code 0

    .. code-block:: python3

        >>> data_for_plot[::30]

    +-----+------------+---------+-----------+-----------+-------+
    |     | iteration  | type    | x         | y         | size  |
    +=====+============+=========+===========+===========+=======+
    | 0   | 0          | middle  | 0.500000  | 0.250000  | 3     |
    +-----+------------+---------+-----------+-----------+-------+
    | 30  | 4          | left    | -0.291796 | 0.085145  | 3     |
    +-----+------------+---------+-----------+-----------+-------+
    | 60  | 8          | right   | 0.042572  | 0.001812  | 3     |
    +-----+------------+---------+-----------+-----------+-------+

    :param history: a history object. a dict with lists. keys iteration, f_value, middle_point, left_point, right_point
    :param func: the functions for which the story was created

    :return: pd.DataFrame for gen_animation_gss. index - num of iteration.
    """

    n = len(history['middle_point'])

    df_middle = pd.DataFrame({'iteration': history['iteration'],
                              'type': ['middle'] * n,
                              'x': history['middle_point'],
                              'y': history['f_value'],
                              'size': [3] * n})  # Data for frames for middle point

    df_left = pd.DataFrame({'iteration': history['iteration'],
                            'type': ['left'] * n,
                            'x': history['left_point'],
                            'y': list(map(func, history['left_point'])),
                            'size': [3] * n})  # Data for frames for left point

    df_right = pd.DataFrame({'iteration': history['iteration'],
                             'type': ['right'] * n,
                             'x': history['right_point'],
                             'y': list(map(func, history['right_point'])),
                             'size': [3] * n})  # Data for frames for right point

    df = pd.concat([df_middle, df_left, df_right]).reset_index(drop=True)  # Concatenate all

    return df


def gen_animation_gss(func: Callable,
                      bounds: tuple[float, float],
                      history: HistoryGSS,
                      **kwargs) -> go.Figure:
    """
    Generates an animation of the golden-section search on `func` between the `bounds`

    :param func: callable that depends on the first positional argument
    :param bounds: tuple with left and right points on the x-axis
    :param history: a history object. a dict with lists. keys iteration, f_value, middle_point, left_point, right_point
    :return: go.Figure with graph

    .. code-block:: python3

        >>> def f(x): return x ** 3 - x ** 2 - x
        >>> _, h = golden_section_search(f, (0, 2), keep_history=True)
        >>> gen_animation_gss(f, (0, 2), h)

    """
    x_axis = np.linspace(bounds[0], bounds[1], 500)
    f_axis = np.zeros_like(x_axis)
    diff_x = max(x_axis) - min(x_axis)

    for i, _x in enumerate(x_axis):
        f_axis[i] = func(_x, **kwargs)
    diff_f = max(f_axis) - min(f_axis)

    df = transfer_history_gss(history, func)
    fig = px.scatter(df, x='x', y='y', size='size', color='type',
                     animation_frame='iteration',
                     range_x=[min(x_axis) - diff_x * 0.15, max(x_axis) + diff_x * 0.15],
                     range_y=[min(f_axis) - diff_f * 0.15, max(f_axis) + diff_f * 0.15],
                     size_max=10,
                     title='<b>Golden section search</b>')

    fig.add_trace(go.Scatter(x=x_axis, y=f_axis, name='function'))

    fig.update_layout(**standard_layout)

    return fig


def gen_animation_spi(func: Callable[[float], float],
                      bounds: [float, float],
                      history: HistorySPI) -> go.Figure:
    """
    Generate animation. Per each iteration we create a go.Frame with parabola plot passing through three points

    :param history: a history object. a dict with lists. keys iteration, f_value, middle_point, left_point, right_point
    :param bounds: tuple with left and right points on the x-axis
    :param func: the functions for which the story was created

    .. code-block:: python3

        >>> def f(x): return x ** 3 - x ** 2 - x
        >>> _, h = successive_parabolic_interpolation(f, (0, 2), keep_history=True)
        >>> gen_animation_spi(f, (0, 2), h)

    """

    n = len(history['iteration'])
    x_axis = torch.linspace(bounds[0], bounds[1], 200)
    f_axis = torch.zeros_like(x_axis)
    diff_x = max(x_axis) - min(x_axis)

    for i, _x in enumerate(x_axis):
        f_axis[i] = func(_x)

    diff_f = max(f_axis) - min(f_axis)
    x_range = [x_axis[0] - diff_x * 0.1, x_axis[-1] + diff_x * 0.1]
    f_range = [min(f_axis) - diff_f * 0.25, max(f_axis) + diff_f * 0.25]

    history = pd.DataFrame(history)
    a, b, c = parabolic_coefficients(
        history.loc[0, 'x0'],
        history.loc[0, 'x1'],
        history.loc[0, 'x2'],
        func
    )

    x0, x1, x2 = history.loc[0, ['x0', 'x1', 'x2']].values

    data = [
        go.Scatter(
            x=x_axis, y=float(a) * x_axis ** 2 + float(b) * x_axis + float(c),
            name='parabola 1', marker={'color': 'rgba(55, 101, 164, 1)'}
        ),
        go.Scatter(
            x=[float(x0)], y=[func(float(x0))],
            name='x0', mode='markers', marker={'size': 10, 'color': 'rgba(66, 122, 161, 1)'}
        ),
        go.Scatter(
            x=[float(x1)], y=[func(float(x1))],
            name='x1', mode='markers', marker={'size': 10, 'color': 'rgba(66, 122, 161, 1)'}
        ),
        go.Scatter(
            x=[float(x2)], y=[func(float(x2))],
            name='x2', mode='markers', marker={'size': 10, 'color': 'rgba(231, 29, 54, 1)'}
        ),
        go.Scatter(
            x=x_axis, y=float(a) * x_axis ** 2 + float(b) * x_axis + float(c),
            name='parabola 0', marker={'color': 'rgba(55, 101, 164, 0.3)'},
        ),
        go.Scatter(
            x=x_axis, y=f_axis, name='function', marker={'color': 'rgba(225, 81, 85, 0.8)'}
        )
    ]

    layout = _make_layout(n, x_range, f_range, 'Successive parabolic search')

    frames = []

    x0, x1, x2 = history.loc[0, ['x0', 'x1', 'x2']].values
    a, b, c = parabolic_coefficients(x0, x1, x2, func)
    parabola_pre = float(a) * x_axis ** 2 + float(b) * x_axis + float(c)
    for i in range(0, history.shape[0]):
        x0, x1, x2 = history.loc[i, ['x0', 'x1', 'x2']].values
        a, b, c = parabolic_coefficients(x0, x1, x2, func)

        parabola_new = float(a) * x_axis ** 2 + float(b) * x_axis + float(c)
        frames.append(go.Frame({
            'data': [
                go.Scatter(
                    x=x_axis, y=parabola_new, name=f'parabola {i + 1}',
                    marker={'color': 'rgba(55, 101, 164, 1)'}
                ),
                go.Scatter(
                    x=[float(x0)], y=[func(float(x0))],
                    name='x0', mode='markers', marker={'size': 10, 'color': 'rgba(66, 122, 161, 1)'}
                ),
                go.Scatter(
                    x=[float(x1)], y=[func(float(x1))],
                    name='x1', mode='markers', marker={'size': 10, 'color': 'rgba(66, 122, 161, 1)'}
                ),
                go.Scatter(
                    x=[float(x2)], y=[func(float(x2))],
                    name='x2', mode='markers', marker={'size': 10, 'color': 'rgba(231, 29, 54, 1)'}
                ),
                go.Scatter(
                    x=x_axis, y=parabola_pre, name=f'parabola {max(i, 1)}',
                    marker={'color': 'rgba(55, 101, 164, 0.3)'}, mode='lines'
                )
            ],

            'name': f'{i}'}))
        parabola_pre = parabola_new

    fig = go.Figure(data=data, layout=layout, frames=frames)
    fig.update_xaxes(range=x_range)
    fig.update_yaxes(range=f_range)
    fig.update_layout(**standard_layout)

    return fig


def parabolic_coefficients(x0: float, x1: float, x2: float, func: Callable[[float], float]) -> [float, float, float]:
    """
    Returns a parabolic function passing through the specified points x0, x1, x2 coeficients

    .. code-block:: python3

        >>> parabolic_coefficients(0, 1, 2, lambda x: x ** 2)
        (1.0, 0.0, 0.0)

    .. math::

        \\begin{bmatrix} a \\\\ b \\\\ c \\end{bmatrix} = \\begin{bmatrix} x_0^2 & x_0 & 1 \\\\ x_1^2 & x_1 & 1 \\\\
        x_2^2 & x_2 & 1 \\end{bmatrix}^{-1} \\cdot \\begin{bmatrix} y_0 \\\\ y_1 \\\\ y_2 \\end{bmatrix}

    :param x0: first point
    :param x1: second point
    :param x2: third point
    :param func: the functions for which the story was created
    :return: coefficients of the parabolic function
    """
    assert x0 != x1 and x0 != x2 and x1 != x2, 'x0, x1, x2 must be different'
    x_mat = torch.tensor([[x0 ** 2, x0, 1.], [x1 ** 2, x1, 1.], [x2 ** 2, x2, 1.]]).double()
    y_mat = torch.tensor([[func(x0)], [func(x1)], [func(x2)]]).double()
    a, b, c = map(float, (torch.linalg.inv(x_mat) @ y_mat).flatten())

    return a, b, c


def _init_function_plot(bounds: list[float, float],
                        func: Callable[[float], float]) -> go.Scatter:
    """
    Returns initial function plot

    :param bounds: left and right constants
    :param func: initial optimization function
    """
    bounds = list(bounds)
    bounds.sort()
    delta = bounds[1] - bounds[0]

    x = torch.linspace(bounds[0] - delta * 0.1, bounds[1] + delta * 0.1, 100)
    fx = list(map(func, x))

    return go.Scatter(x=x, mode='lines', y=fx, name='function', marker={'size': 5, 'color': COLOR1})


def _make_golden_frame(x: list[float, float, float],
                       func: Callable[[float], float],
                       func_plot: go.Scatter,
                       i: int) -> go.Frame:
    """
    Returns golden step frame. Also, first step (initialization)

    :param x: list of x. x[0] - left bound, x[1] - middle, x[2] - right bound
    :param func: initial function
    :param func_plot: go.Scatter of initial function
    :param i: frame's number
    :return: one frame of golden section search step
    """
    x = sorted(list(x), key=func, reverse=True)
    fx = list(map(func, x))

    data = [
        go.Scatter(x=x[:1], y=fx[:1], name='x2', marker={'size': 10, 'color': COLOR3}, mode='markers'),
        go.Scatter(x=x[1:2], y=fx[1:2], name='x1', marker={'size': 10, 'color': COLOR3}, mode='markers'),
        go.Scatter(x=x[2:3], y=fx[2:3], name='x0', marker={'size': 10, 'color': COLOR5}, mode='markers'),
        func_plot,
        go.Scatter(x=[0], y=[0], name='golden step', mode='lines', marker={'size': 10, 'color': 'rgba(0, 0, 0, 0)'})
    ]
    return go.Frame(data=data, name=f'{i}')


def _make_parabolic_frame(x: list[float, float, float],
                          func: Callable[[float], float],
                          func_plot: go.Scatter,
                          i: int) -> go.Frame:
    """
    Returns successive step frame.

    :param x: list of x. x[0] - left bound, x[1] - middle, x[2] - right bound
    :param func: initial function
    :param func_plot: go.Scatter of initial function
    :param i: frame's number
    :return: one frame of successive parabolic step
    """
    a, b, c = parabolic_coefficients(*x, func=func)
    x = sorted(list(x), key=func, reverse=True)
    fx = list(map(func, x))
    x_axis = func_plot.x
    y_axis = a * x_axis ** 2 + b * x_axis + c

    data = [
        go.Scatter(x=x[:1], y=fx[:1], name='x2', marker={'size': 10, 'color': COLOR3}, mode='markers'),
        go.Scatter(x=x[1:2], y=fx[1:2], name='x1', marker={'size': 10, 'color': COLOR3}, mode='markers'),
        go.Scatter(x=x[2:3], y=fx[2:3], name='x0', marker={'size': 10, 'color': COLOR5}, mode='markers'),
        func_plot,
        go.Scatter(
            x=x_axis, y=y_axis, name='parabolic step',
            marker={'size': 8, 'color': COLOR2}, mode='lines', showlegend=True
        ),
    ]
    return go.Frame(data=data, name=f'{i}')


def _make_frames(history: HistoryBrent,
                 func: Callable[[float], float]) -> list:
    """
    Returns frames of brent optimization

    :param history: History of optimization by brent algorithm
    :param func: initial function
    :return: frames of brent optimization
    """
    hdf = pd.DataFrame(history)

    initial_function = _init_function_plot(hdf.loc[0, ['left_bound', 'right_bound']], func)
    frames = []

    for i in range(1, hdf.shape[0]):
        if hdf.loc[i, 'type_step'] == 'golden':
            frame = _make_golden_frame(
                hdf.loc[i, ['left_bound', 'x_least', 'right_bound']],
                func,
                initial_function,
                i
            )
        else:
            frame = _make_parabolic_frame(
                hdf.loc[i, ['x_least', 'x_middle', 'x_largest']],
                func,
                initial_function,
                i
            )

        frames.append(frame)

    return frames


def _make_layout(n: int, x_range: list[float, float], f_range: list[float, float], title: str) -> go.Layout:
    """
    Returns go.Layout for spi and brent algorithms

    :param n: number of frames
    :param x_range: constraints of x
    :param f_range: constraints of y (second axi)
    :param title: title of layout
    :return: go layout with sliders
    """
    layout = go.Layout(
        {
            'font': {'size': 14},
            'legend': {'title': {'text': '<b>variable names</b>'}},
            'sliders': [
                {
                    'active': 0,
                    'currentvalue': {'prefix': 'iteration='},
                    'len': 0.9,
                    'pad': {'b': 10, 't': 60},
                    'steps': [
                        {
                            'args': [
                                [f'{i}'],
                                {
                                    'frame': {'duration': 500, 'redraw': False, 'mode': 'immediate'},
                                    'mode': 'immediate',
                                    'fromcurrent': False,
                                    'transition': {'duration': 300, 'easing': 'cubic-in-out'}
                                }
                            ],
                            'label': f'{i}',
                            'method': 'animate'
                        } for i in range(n)
                    ],
                    'x': 0.1,
                    'xanchor': 'left',
                    'y': 0,
                    'yanchor': 'top'
                }
            ],

            'title': {'text': f'<b>{title}</b>'},
            'updatemenus': [
                {'buttons': [
                    {
                        'args': [
                            None,
                            {
                                'frame': {'duration': 1000, 'redraw': True},
                                'mode': 'immediate',
                                'fromcurrent': True,
                                'transition': {'duration': 300, 'easing': 'cubic-in-out'}
                            }
                        ],
                        'label': '&#9654;',
                        'method': 'animate'
                    },
                    {
                        'args': [
                            [None],
                            {
                                'frame': {'duration': 100, 'redraw': True},
                                'mode': 'immediate',
                                'fromcurrent': True,
                                'transition': {'duration': 100, 'easing': 'cubic-in-out'}
                            }
                        ],
                        'label': '&#9724;',
                        'method': 'animate'
                    }
                ],
                    'direction': 'left',
                    'pad': {'r': 10, 't': 70},
                    'showactive': True,
                    'type': 'buttons',
                    'x': 0.1,
                    'xanchor': 'right',
                    'y': 0,
                    'yanchor': 'top'
                }
            ],

            'xaxis': {
                'anchor': 'y',
                'domain': [0.0, 1.0],
                'range': x_range,
                'autorange': False,
                'title': {'text': '<b>x</b>'}
            },
            'yaxis': {
                'anchor': 'x',
                'domain': [0.0, 1.0],
                'range': f_range,
                'autorange': False,
                'title': {'text': '<b>f(x)</b>'}
            }
        }
    )
    return layout


def gen_animation_brent(func: Callable[[float], float], history: HistoryBrent) -> go.Figure:
    """
    Returns a visualization of the Brent algorithm. Each iteration shows which iteration.

    :param func: callable that depends on the first positional argument
    :param history: brent optimization history
    :return: animation of optimization

    .. code-block:: python3

        >>> def f(x): return x ** 3 - x ** 2 - x
        >>> _, h = brent(f, (0, 2), keep_history=True)
        >>> gen_animation_brent(f, h)

    """

    # x bounds
    lb, rb, xl = history['left_bound'][0], history['right_bound'][0], history['x_least'][0]
    delta = abs(rb - lb)
    x_range = [lb - delta * 0.2, rb + delta * 0.2]

    # initial frame and data
    hdf = pd.DataFrame(history)
    initial_function = _init_function_plot([lb, rb], func)
    data = _make_golden_frame([lb, xl, rb], func, initial_function, 0).data

    # all frames
    frames = _make_frames(history, func)

    # y bounds
    y_bottom, y_top = min(initial_function.y), max(initial_function.y)
    delta = y_top - y_bottom
    y_range = [y_bottom - delta * 0.25, y_top + delta * 0.25]

    # layout
    layout = _make_layout(hdf.shape[0], x_range, y_range, 'Brent optimization')

    # all visualization
    return go.Figure(data=data, layout=layout, frames=frames)



from __future__ import annotations

import sys
from typing import Dict, Union, Optional

import plotly.express as px
import plotly.graph_objects as go

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal
import torch


def check_tensors(*tensors: torch.Tensor) -> [torch.Tensor]:
    """
    Returns flattened tensors and checks if the arrays have the same size and flare.Tensor

    :param tensors: tensors
    :return: tensors
    """
    shapes = []
    output = []
    for arr in tensors:
        if arr is not None:
            assert isinstance(arr, torch.Tensor), 'arrays must be torch Tensor'
            arr = arr.flatten()
            shapes.append(int(arr.shape[0]))

        output.append(arr)

    assert len(torch.unique(torch.tensor(shapes))) == 1, 'Tensors must have same sizes'

    return output


def tpr(y_true: torch.Tensor, y_pred: torch.Tensor) -> float:
    """
    Return True Positive Rate. TPR = TP / P = TP / (TP + FN). Alias is Recall

    .. note:: if P == 0, then TPR = 0

    :param y_true: array with true values of binary classification
    :param y_pred: array with prediction values of binary classification
    :return:
    """
    y_true, y_pred = check_tensors(y_true, y_pred)
    tp = ((y_true == y_pred) & (y_true == 1)).sum()
    p = (y_true == 1).sum()
    return float(tp / max(p, 1))


def fpr(y_true: torch.Tensor, y_pred: torch.Tensor) -> float:
    """
    Return False Positive Rate. FPR = FP / N = FP / (FP + TN).

    .. note:: if N == 0, then FPR = 0

    :param y_true: array with true values of binary classification
    :param y_pred: array with prediction values of binary classification
    :return:
    """
    y_true, y_pred = check_tensors(y_true, y_pred)
    fp = ((y_true != y_pred) & (y_true == 0)).sum()
    n = (y_true == 0).sum()
    return float(fp / max(n, 1))


def precision(y_true: torch.Tensor, y_pred: torch.Tensor) -> float:
    """
    Return Positive Predictive Value . PPV = TP / (TP + FP)

    .. note:: if TP + FP == 0, then PPV = 0

    :param y_true: array with true values of binary classification
    :param y_pred: array with prediction values of binary classification
    :return:
    """
    y_true, y_pred = check_tensors(y_true, y_pred)
    tp = ((y_true == y_pred) & (y_true == 1)).sum()
    fp = ((y_true != y_pred) & (y_true == 0)).sum()
    return float(tp / max(tp + fp, 1))


recall = tpr


def accuracy(y_true: torch.Tensor, y_pred: torch.Tensor) -> float:
    """
    Return accuracy. ACC = (TP + TN) / (P + N) = (TP + TN) / (TP + FP + TN + FN)

    :param y_true: array with true values of binary classification
    :param y_pred: array with prediction values of binary classification
    :return:
    """
    y_true, y_pred = check_tensors(y_true, y_pred)
    return float(torch.eq(y_true, y_pred).sum() / y_true.shape[0])


def f_score(y_true: torch.Tensor, y_pred: torch.Tensor, beta: float = 1) -> float:
    """
    Return F_score. https://en.wikipedia.org/wiki/F-score

    .. math::
        F_\\beta = (1 + \\beta^2) \\cdot \\frac{\\mathrm{precision} \\cdot \\mathrm{recall}}
        {(\\beta^2 \\cdot \\mathrm{precision}) + \\mathrm{recall}}.

    .. note:: if beta ** 2 * _precision + _recall == 0, then f_score = 0

    :param y_true: array with true values of binary classification
    :param y_pred: array with prediction values of binary classification
    :param beta: is chosen such that recall is considered beta times as important as precision
    :return:
    """
    y_true, y_pred = check_tensors(y_true, y_pred)
    _precision = precision(y_true, y_pred)
    _recall = tpr(y_true, y_pred)

    numerator = (1 + beta ** 2) * (_precision * _recall)
    denominator = max(beta ** 2 * _precision + _recall, 1e-12)

    return float(numerator / denominator)


def best_threshold(x: torch.Tensor, y_true: torch.Tensor, model: torch.nn.Module,
                   metric: Literal['f1', 'by_roc'] = 'f1', step_size: float = 0.01):
    """
    Returns best threshold by metric by linear search

    :param x: training tensor
    :param y_true: target tensor. array with true values of binary classification
    :param model: some model that returns a torch tensor with class 1 probabilities using the call: model(x)
    :param metric: name of the target metric that we need to maximize. by_roc - difference between TPR and FPR
    :param step_size: step size of linear search
    :return:
    """
    y_true = y_true.flatten()
    metric = {'f1': f_score, 'by_roc': lambda y1, y2: tpr(y1, y2) - fpr(y1, y2)}[metric]
    best_t = 0
    best_metric = 0
    y_prob = model(x).flatten()

    for threshold in torch.arange(0, 1 + step_size, step_size):

        y_pred = (y_prob >= threshold) * 1
        metric_i = metric(y_true, y_pred)

        if metric_i > best_metric:
            best_metric = metric_i
            best_t = threshold

    return best_t


def roc_curve(y_true: torch.Tensor, y_prob: torch.Tensor, n_thresholds: Union[int, None] = None) -> Dict:
    """
    Return dict with points at TPR - FPR coordinates

    :param y_true: array with true values of binary classification
    :param y_prob: array of probabilities of confidence of belonging to the 1st class
    :param n_thresholds: if len(y_true) is too large, you can limit the number of threshold values
    :return: dict with values of TPR and FPR
    """
    tpr_array = []
    fpr_array = []
    check_tensors(y_true, y_prob)

    thresholds, _ = torch.sort(torch.unique(y_prob), descending=True)
    thresholds = torch.cat((torch.ones(1), thresholds, torch.zeros(1)))

    if n_thresholds is not None:
        thresholds = thresholds[torch.linspace(0, len(thresholds) - 1, n_thresholds, dtype=torch.long)]

    for threshold in thresholds:
        tpr_array.append(tpr(y_true, (y_prob >= threshold) * 1))
        fpr_array.append(fpr(y_true, (y_prob >= threshold) * 1))

    return {'TPR': tpr_array, 'FPR': fpr_array}


def auc_roc(y_true: torch.Tensor, y_prob: torch.Tensor, n_thresholds: int = 500) -> float:
    """
    Return area under curve ROC (AUC-ROC metric)

    :param y_true: array with true values of binary classification
    :param y_prob: array of probabilities of confidence of belonging to the 1st class
    :param n_thresholds: if len(y_true) is too large, you can limit the number of threshold values
    :return: float value of area under roc-curve
    """
    check_tensors(y_true, y_prob)

    tpr_array, fpr_array = roc_curve(y_true, y_prob, min(y_true.shape[0], n_thresholds)).values()
    auc = 0
    for i in range(len(fpr_array) - 1):  # Integrating
        auc += tpr_array[i + 1] * (fpr_array[i + 1] - fpr_array[i])
    return float(auc)


def binary_classification_report(y_true: torch.Tensor,
                                 y_pred: torch.Tensor,
                                 y_prob: Optional[torch.Tensor] = None,
                                 ) -> dict:
    """
    Returns dict with recall, precision, accuracy, f1, auc roc scores

    :param y_true: array with true values of binary classification
    :param y_pred: array with prediction values of binary classification
    :param y_prob: array of probabilities of confidence of belonging to the 1st class
    :return: float value of area under roc-curve
    """
    check_tensors(y_true, y_pred, y_prob)

    return {
        'recall': recall(y_true, y_pred),
        'precision': precision(y_true, y_pred),
        'f1': f_score(y_true, y_pred),
        'auc_roc': None if y_prob is None else auc_roc(y_true, y_prob)
    }


def r2_score(y_true: torch.Tensor, y_pred: torch.Tensor) -> float:
    """
    Return R2 metric of regression

    .. math::

        \\mathbf{R}^{2} = 1 - \\frac{\\sum_{i=1}^{n}(\\hat y_i - y_i)^2}{\\sum_{i = 1}^{n}(\\hat y_i - \\overline y)^2}

    :param y_true: array with true values of binary classification
    :param y_pred: array with prediction values of binary classification
    :return:
    """
    y_true, y_pred = check_tensors(y_true, y_pred)
    y_true, y_pred = y_true.float(), y_pred.float()
    rss = (y_true - y_pred).norm(2) ** 2
    tss = (y_true - y_true.mean()).norm(2) ** 2

    return float(1 - rss / tss)


def roc_curve_plot(y_true: torch.Tensor, y_prob: torch.Tensor, fill: bool = False) -> go.Figure:
    """
    Return figure with plotly.Figure ROC curve

    :param y_true: array with true values of binary classification
    :param y_prob: array of probabilities of confidence of belonging to the 1st class
    :param fill: flag for filling the area under the curve
    :return: go.Figure

    .. code-block:: python3

        >>> yt = torch.tensor([1, 1, 0, 0, 1, 0])
        >>> yp = torch.tensor([0.7, 0.6, 0.3, 0.5, 0.4, 0.4])
        >>> roc_curve_plot(yt, yp)

    """
    if fill:
        fig = px.area(roc_curve(y_true, y_prob, None if len(y_true) < 1000 else 1000), x='FPR', y='TPR',
                      title='<b>ROC curve</b>')
    else:
        fig = px.line(roc_curve(y_true, y_prob, None if len(y_true) < 1000 else 1000), x='FPR', y='TPR',
                      title='<b>ROC curve</b>')

    fig.update_layout(font={'size': 18}, autosize=False, width=700, height=600, xaxis={'range': [-0.05, 1.05]})
    fig.add_scatter(x=[0, 1], y=[0, 1], mode='lines', line={'dash': 'dash'}, name='', showlegend=False)
    return fig



from __future__ import annotations

import sys
from typing import Callable

import numpy as np
import torch

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from .metrics import binary_classification_report, best_threshold, f_score


class BaseClassification(torch.nn.Module):

    def __init__(self):
        super(BaseClassification, self).__init__()

    def metrics_tab(self, x: torch.Tensor, y: torch.Tensor) -> dict:
        """
        Returns metrics dict with recall, precision, accuracy, f1, auc roc scores

        :param x: training set
        :param y: target value
        :param: dict with recall, precision, accuracy, f1, auc roc scores
        :return:
        """
        y_prob: torch.Tensor = self.forward(x)
        try:
            y_pred: torch.Tensor = (y_prob > self.best_threshold) * 1
        except AttributeError:
            y_pred = self.predict(x)
            y_prob: None = None

        return binary_classification_report(y, y_pred, y_prob)


class LogisticRegression(BaseClassification):
    """
    Binary classification model

    Let :math:`x \\in \\mathbb{R}^{n \\times m}, \\ w \\in \\mathbb{R}^{m \\times 1}, \\ I = [1]_{n \\times 1}`,
    :math:`x_i` -- is a row and :math:`x_i \\in \\mathbb{R}^{1 \\times m}`

    Model:

    .. math::

        \\mathbb{P}(y_i = 1 | w) = \\frac{1}{1 + \\exp (x_i \\cdot w + b)}


    """

    def __init__(self, kernel: Literal['linear', 'perceptron'] = 'linear'):
        """
        :param kernel: 'linear' or 'perceptron'. linear - basic logistic regression, perceptron - nn with 2
                       hidden layer with dim1 = 1024, dim2 = 512
        """
        super(LogisticRegression, self).__init__()

        self.sigmoid = torch.nn.Sigmoid()
        self.weights = None
        self.kernel = kernel
        if kernel not in ['linear', 'perceptron']:
            raise TypeError('Invalid kernel. Choose "linear" or "perceptron"')

        self.best_threshold = 0.5

    def init_weights(self, x: torch.Tensor):
        """
        Initialization weights

        :param x: input torch tensor
        """
        if self.kernel == 'linear':
            self.weights = torch.nn.Linear(x.shape[1], 1)
        elif self.kernel == 'perceptron':
            self.weights = torch.nn.Sequential(
                torch.nn.Linear(x.shape[1], 1024),
                torch.nn.ReLU(),
                torch.nn.Linear(1024, 512),
                torch.nn.ReLU(),
                torch.nn.Linear(512, 1)
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Returns confidence probabilities of first class

        :param x: training set
        :return: probabilities
        """

        if self.weights is None:
            self.init_weights(x)

        x = x.float()
        return self.sigmoid(self.weights(x))

    def predict(self, x):
        """
        Returns binary class 0 or 1 instead of probabilities

        :param x: some tensor with shape[1] = n_features
        :return:
        """
        y_prob: torch.Tensor = self.forward(x).flatten()
        y_pred: torch.Tensor = (y_prob > self.best_threshold) * 1
        return y_pred

    def fit(self, x: torch.Tensor, y: torch.Tensor, epochs=1000, l1_lambda: float = 0,
            show_epoch: int = 0, print_function: Callable = print) -> torch.nn.Module:
        """
        Returns trained model Logistic Regression

        :param x: training set
        :param y: target value
        :param epochs: max number of sgd implements
        :param l1_lambda: l1 regularization weight
        :param show_epoch: amount of showing epochs
        :param print_function: print or streamlit.write
        :return: trained model
        """
        x = x.float()
        y = y.float()

        self.forward(x)

        print_epochs = np.unique(np.linspace(1, epochs, min(epochs, show_epoch), dtype=int))

        optimizer = torch.optim.Adam(self.parameters())
        loss = torch.nn.BCELoss()

        for epoch in range(1, epochs + 1):
            optimizer.zero_grad()
            output = loss(self.forward(x).flatten(), y.flatten())
            if l1_lambda > 0:
                for layer in self.parameters():
                    output += l1_lambda * layer.data.sum()

            output.backward()
            optimizer.step()

            with torch.no_grad():
                if epoch in print_epochs:
                    print_function(f'Epoch: {epoch: 5d} | CrossEntropyLoss: {output.item(): 0.5f}')

        self.best_threshold = best_threshold(x, y, self, metric='f1')

        return self


class LogisticRegressionRBF(BaseClassification):
    """

    This is a logistic regression, but before we make a basic linear prediction and apply the sigmoid, we transfer x to
    another space using radial basis functions. The dimension of this space depends on the basis matrix x (x_basis) [1]_

    .. rubric:: Radial basis functions

    #. gaussian :math:`\\displaystyle \\varphi (x, x_b)=e^{-\\Vert x - x_b \\Vert^2}`

    #. linear :math:`\\varphi (x, x_b) = \\Vert x - x_b \\Vert`

    #. multiquadratic :math:`\\displaystyle \\varphi (x, x_b)  = \\sqrt{1 + \\Vert x - x_b \\Vert^2}`

    .. rubric:: References

    .. [1] https://en.wikipedia.org/wiki/Radial_basis_function

    """

    def __init__(self, x_basis: torch.Tensor, rbf: Literal['linear', 'gaussian', 'multiquadratic'] = 'gaussian'):
        """

        :param x_basis: centers of basis functions
        :param rbf: type of rbf function. Available: ['linear', 'gaussian']
        """

        super(LogisticRegressionRBF, self).__init__()

        self.w = torch.nn.Linear(x_basis.shape[0], 1)
        self.rbf = rbf
        self.x_basis = x_basis
        self.sigmoid = torch.nn.Sigmoid()
        self.best_threshold = 0.5

    def forward(self, x: torch.Tensor = None, phi_matrix: torch.Tensor = None) -> torch.Tensor:
        """
        Returns a "probability" (confidence) of class 1

        :param x: 2D array
        :param phi_matrix: 2D array
        :return: 1D array
        """
        if phi_matrix is None:
            phi_matrix = self.make_phi_matrix(x)

        return self.sigmoid(self.w(phi_matrix))

    def make_phi_matrix(self, x: torch.Tensor) -> torch.Tensor:
        """
        Returns n x k array with calculated phi(x_i, x_basis_j). n is number of observation from x (x.shape[0])
        k is number of basis from initialization.

        .. math::
            
            \\begin{bmatrix} \\varphi(x_1, x^\\text{basis}_1) & \\varphi(x_1, x^\\text{basis}_2) & \\dots &
            \\varphi(x_1, x^\\text{basis}_k) \\\\ \\varphi(x_2, x^\\text{basis}_1) & \\varphi(x_2, x^\\text{basis}_2) &
            \\dots &  \\varphi(x_2, x^\\text{basis}_k) \\\\ \\vdots & \\vdots & \\ddots & \\vdots \\\\
            \\varphi(x_n, x^\\text{basis}_1) & \\varphi(x_n, x^\\text{basis}_2) & \\dots &
            \\varphi(x_n, x^\\text{basis}_k) \\ \\end{bmatrix}

        :param x: Array k x m dimensional. k different x_i and m features
        """
        x = x.float()
        n = self.x_basis.shape[0]
        k = x.shape[0]

        repeated_input_x = torch.tile(x, (n, 1))
        repeated_basis_x = torch.tile(self.x_basis, (1, k))
        repeated_basis_x = torch.reshape(repeated_basis_x, repeated_input_x.shape)

        phi = ((repeated_input_x - repeated_basis_x) ** 2).sum(dim=1)
        phi = torch.reshape(phi, (n, k)).T

        if self.rbf == 'linear':
            phi = phi ** 0.5
            phi = phi / phi.max()

        elif self.rbf == 'gaussian':
            phi = torch.exp(-phi)

        elif self.rbf == 'multiquadratic':
            phi = (1 + phi) ** 0.5
            phi = phi / phi.max()

        return phi.float()

    def predict(self, x):
        """
        Returns binary class 0 or 1 instead of -1; 1

        :param x: some tensor with shape[1] = n_features
        :return:
        """
        y_prob: torch.Tensor = self.forward(x).flatten()
        y_pred: torch.Tensor = (y_prob > self.best_threshold) * 1
        return y_pred

    def fit(self,
            x: torch.Tensor,
            y: torch.Tensor,
            epochs=100,
            l1_lambda: float = 0,
            show_epoch: int = 0,
            print_function: Callable = print) -> torch.nn.Module:
        """
        Returns trained model Logistic Regression with RBF

        :param x: training set
        :param y: target value
        :param epochs: max number of sgd implements
        :param l1_lambda: l1 regularization weight
        :param show_epoch: amount of showing epochs
        :param print_function: e.g. print or streamlit.write
        :return: trained model
        """
        x = x.float()
        y = y.float()

        print_epochs = np.unique(np.linspace(1, epochs, min(epochs, show_epoch), dtype=int))

        phi_matrix = self.make_phi_matrix(x)
        optimizer = torch.optim.Adam(self.parameters())
        loss = torch.nn.BCELoss()

        for epoch in range(1, epochs + 1):
            optimizer.zero_grad()
            output = loss(self.forward(x, phi_matrix).flatten(), y.flatten())
            if l1_lambda > 0.:
                for layer in self.parameters():
                    output += l1_lambda * layer.data.sum()

            output.backward()
            optimizer.step()

            with torch.no_grad():
                if epoch in print_epochs:
                    print_function(f'Epoch: {epoch: 5d} | CrossEntropyLoss: {output.item(): 0.5f}')

        self.best_threshold = best_threshold(x, y, self, metric='f1')

        return self


class SVM(BaseClassification):
    """

    Binary classification model. Method predict: SVM.predict(x) --> original names

    Mathematical model:

    .. math::

        \\hat y = \\operatorname{sign}(x \\cdot w - b \\cdot I)


    :math:`x \\in \\mathbb{R}^{n \\times m}, \\ w \\in \\mathbb{R}^{m \\times 1}, \\ I = [1]_{n \\times 1}`


    And search of best :math:`w, b` calculates by minimization of Hinge loss

    .. math::

        {\\displaystyle \\lambda \\lVert \\mathbf {w} \\rVert ^{2}+\\left[{\\frac {1}{n}}\\sum _{i=1}^{n}\\max
        \\left(0,1-y_{i}(x_i \\cdot w - b)\\right)\\right] \\longrightarrow \\min }


    or PEGASOS algorithm

    :ivar scale: for the best training and prediction, the model will standard normalize the input x data.
                 The first time you call model, std and mean will be saved and in the future use the parameters
                 for scaling. x = (x is the average value) / std
    :ivar weights: parameters of model. Initialize after first calling

    """

    def __init__(self):
        """
        Initialization of SVM
        """
        super(SVM, self).__init__()
        self.scale = None
        self.weights = None
        self.class_names = {0: -1, 1: 1}  # the null class has the name -1, and the first class has the name 1

    def init_weights(self, x: torch.Tensor):
        """
        Initialization weights

        :param x: input torch tensor
        """
        self.weights = torch.nn.Linear(x.shape[1], 1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Returns x @ w + b

        .. math::
            f(x) = w_0 + w_1 \\cdot x_1 + w_2 \\cdot x_2 + \\dots + w_m \\cdot x_m

        :param x: input observations, tensor n x m (n is the number of observations that have m parameters)
        :return: regression value (yes, no classification, for binary classes call predict)
        """
        x = x.float()

        if self.weights is None:
            self.init_weights(x)

        return self.weights(x)

    def scaler(self, x: torch.Tensor) -> torch.Tensor:
        """
        Returns the scaled value of x. Standard x scaling and storing settings

        :param x: torch.Tensor
        :return:
        """
        if self.scale is None:
            self.scale = (x.mean(), x.flatten().std())

        return (x - self.scale[0]) / self.scale[1]

    def fit(self,
            x: torch.Tensor,
            y: torch.Tensor,
            method: Literal['pegasos', 'sgd'] = 'sgd',
            epochs=100,
            lambda_reg: float = 0.1,
            show_epoch: int = 0,
            print_function: Callable = print):
        """
        Returns trained model SVM

        :param x: training set
        :param y: target value. binary classes
        :param method: optimization method. Available PEGASOS or sgd
        :param epochs: max number of sgd and pegasos steps 
        :param lambda_reg: l2 regularization weight
        :param show_epoch: amount of showing epochs
        :param print_function: print or streamlit.write
        :return: trained model
        """
        y = y.flatten().float().round().int()
        uniq_y = torch.unique(y)
        assert torch.unique(y).shape[0] <= 2, 'binary classification support only'
        if 1 in uniq_y:
            self.class_names = {int(uniq_y[uniq_y != 1][0]): -1, int(uniq_y[uniq_y == 1][0]): 1}
        else:
            self.class_names = dict(zip(map(int, uniq_y), (-1, 1)))

        y_replaced = torch.zeros_like(y)
        for i in range(y_replaced.shape[0]):
            y_replaced[i] = self.class_names[int(y[i])]

        y = y_replaced.float()

        if method == 'pegasos':
            return self._fit_pegasos(x, y, epochs, lambda_reg, show_epoch, print_function)
        else:
            return self._fit_sgd(x, y, epochs, lambda_reg, show_epoch, print_function)

    def _fit_sgd(self, x: torch.Tensor, y: torch.Tensor, epochs=500, l2_lambda: float = 0,
                 show_epoch: int = 0, print_function: Callable = print):
        """
        Returns trained model SVM

        :param x: training set
        :param y: target value
        :param epochs: max number of sgd implements
        :param l2_lambda: l2 regularization weight
        :param show_epoch: amount of showing epochs
        :param print_function: print or streamlit.write
        :return: trained model
        """

        x = self.scaler(x.float())
        self.forward(x)

        print_epochs = np.unique(np.linspace(1, epochs, min(epochs, show_epoch), dtype=int))

        optimizer = torch.optim.Adam(self.parameters(), lr=1e-2, weight_decay=l2_lambda)
        loss = torch.nn.MarginRankingLoss(margin=1)  # hinge loss if x2 = 0 and margin = 1

        for epoch in range(1, epochs + 1):
            optimizer.zero_grad()
            output = loss(self.forward(x).flatten(), torch.tensor([0]), y.flatten())
            output.backward()
            optimizer.step()

            with torch.no_grad():
                if epoch in print_epochs:
                    print_function(f'Epoch: {epoch: 5d} | HingeLoss: {output.item(): 0.5f}')

        return self

    def _fit_pegasos(self, x: torch.Tensor, y: torch.Tensor, epochs=20, lambda_reg: float = 0.95,
                     show_epoch: int = 0, print_function: Callable = print) -> torch.nn.Module:
        """
        Returns trained model SVM [2]_

        :param x: training set
        :param y: target value
        :param epochs: max number of sgd implements
        :param lambda_reg: regularization parameter
        :param show_epoch: amount of showing epochs
        :param print_function: print or streamlit.write
        :return: trained model

        .. rubric:: References

        .. [2] Pegasos: Primal Estimated sub-GrAdient SOlver for SVM. Shai Shalev-Shwartz; Yoram Singer; Nathan Srebro;
               Andrew Cotter

        """
        x = self.scaler(x.float())
        self.forward(x)

        print_epochs = np.unique(np.linspace(1, epochs, min(epochs, show_epoch), dtype=int))
        weights = torch.zeros(x.shape[1])
        bias = torch.zeros(1)
        t = 0
        for epoch in range(1, epochs + 1):
            t += 1
            eta = 1 / (lambda_reg * t)
            for j in torch.randint(0, x.shape[0], (min(100, x.shape[0]),)):

                if y[j] * (weights @ x[j] + bias) < 1:
                    weights = (1 - eta * lambda_reg) * weights + eta * y[j] * x[j]
                    bias = (1 - eta * lambda_reg) * bias + eta * y[j] * 1

                else:
                    weights = (1 - eta * lambda_reg) * weights
                    bias = (1 - eta * lambda_reg) * bias

                weights = min(1, 1 / lambda_reg ** 0.5 / torch.concat([weights, bias]).norm(2)) * weights

                with torch.no_grad():
                    if epoch in print_epochs:
                        print_function(f'Epoch: {epoch: 5d} | F1 score: {f_score(y, self.predict(x)): 0.5f}')

        params = list(self.weights.parameters())
        params[0].data = weights.reshape(1, -1)
        params[1].data = bias

        return self

    def predict(self, x):
        """
        Returns binary class from the first call, in training or just call

        :param x: some tensor with shape[1] = n_features
        :return:
        """
        x = self.scaler(x).float()
        y = torch.zeros(x.shape[0])
        reversed_names = {**{0: 1}, **{v: u for u, v in self.class_names.items()}}
        for i, xi in enumerate(self.weights(x).sign().flatten().int()):
            y[i] = reversed_names[int(xi)]

        return y


if __name__ == '__main__':
    # from sklearn.datasets import make_blobs
    # from metrics import accuracy

    # torch.random.manual_seed(7)
    # _x, _y = make_blobs(1000, centers=2, random_state=8)
    # _x, _y = torch.tensor(_x), torch.tensor(_y)
    # m = SVM().fit(_x, _y)
    # print(accuracy(_y, m.predict(_x)))
    # m = SVM().fit(_x, _y, method='pegasos')
    # print(accuracy(_y, m.predict(_x)))
    # m = LogisticRegression().fit(_x, _y)
    # print(accuracy(_y, m.predict(_x)))
    # m = LogisticRegressionRBF(_x[:100]).fit(_x, _y)
    # print(accuracy(_y, m.predict(_x)))

    from sklearn.datasets import make_moons

    _x, _y = make_moons(10_000, noise=.1, random_state=84)
    _x, _y = torch.tensor(_x), torch.tensor(_y)

    logistic_model_rbf = LogisticRegressionRBF(_x[:1000]).fit(_x, _y, show_epoch=10)

    print(logistic_model_rbf.metrics_tab(_x, _y))



from .calculus import hessian, jacobian, gradient
from .ml import metrics
from .ml.classification import *
from .ml.optimize import *
from .multi_optimize import gd_optimal, gd_frac, gd_constant, nonlinear_cgm, bfgs, primal_dual_interior, \
    log_barrier_solver, constrained_lagrangian_solver
from .one_optimize import golden_section_search, successive_parabolic_interpolation, brent
from .support import HiddenPrints

__all__ = [
    "hessian",
    "jacobian",
    "gradient",
    "golden_section_search",
    "successive_parabolic_interpolation",
    "brent",
    "bfgs",
    "HiddenPrints",
    "gd_constant",
    "gd_frac",
    "gd_optimal",
    "nonlinear_cgm",
    "primal_dual_interior",
    "log_barrier_solver",
    "constrained_lagrangian_solver",
    "ml",
    "metrics",
    "visualize"
]

__version__ = '1.0 dev 2022 ))) Hi'



from __future__ import annotations

from typing import Optional

import plotly.express as px
import plotly.graph_objs as go
import torch
from sklearn.manifold import TSNE

from .one_animation import standard_layout
from ..ml.classification import LogisticRegressionRBF
from ..ml.metrics import roc_curve_plot


def gen_classification_plot(x_tensor: torch.Tensor,
                            y_true: torch.Tensor,
                            model: Optional[torch.nn.Module] = None,
                            threshold: float = 0.5,
                            cnt_points: int = 1000,
                            k: float = 0.1,
                            title: Optional[str] = None,
                            epsilon: float = 1e-4,
                            insert_na: bool = False) -> go.Figure:
    """
    Returns a graph with a distribution and an optional line. If dim(x) = 2, then you can get model. If dim(x) > 2,
    then returns graph of TSNE from sklearn.manifold with default settings. dim(x) is not support

    .. note::

        if model os linear and have one layer, simple activation function, then visualization will faster

    .. warning::

        if the model is heavy, then you should reduce cnt_points, but the probability of missing points is higher,
        and the visualization will be rather incorrect. You can increase the gap by increasing the epsilon.

    :param x_tensor: training tensor
    :param y_true: target tensor. array with true values of binary classification
    :param model: some model that returns a torch tensor with class 1 probabilities using the call: model(x)
    :param threshold: if model(xi) >= threshold, then yi = 1
    :param cnt_points: number of points on each of the two axes when dim(x) = 2
    :param k: constant for draw on section: [x.min() - (x.max() - x.min()) * k, x.max() + (x.max() - x.min()) * k]
    :param title: title of plots
    :param epsilon: contour line points: :math:`\\{x\\in \\mathbb{R}^2 \\, | \\,
                \\text{threshold} - \\text{epsilon} \\le \\text{model}(x) \\le \\text{threshold} + \\text{epsilon}\\}`
    :param insert_na: na insertion flag when two points too far away
    :return: scatter plot go.Figure

    .. code-block:: python3

        >>> from sklearn.datasets import make_moons
        >>> torch.random.manual_seed(7)
        >>> x, y = make_moons(1000, noise=0.15, random_state=7)
        >>> x, y = torch.tensor(x), torch.tensor(y)

        >>> lr_rbf = LogisticRegressionRBF(x[:50])
        >>> lr_rbf.fit(x, y, epochs=5000)

        >>> lr_rbf.metrics_tab(x, y)

    .. code-block:: python3

        {'recall': 0.9980000257492065,
         'precision': 0.9842209219932556,
         'f1': 0.9910625822119956,
         'auc_roc': 0.9995800006320514}

    .. code-block:: python3

        >>> gen_classification_plot(x, y, model, threshold=0.5, epsilon=0.001)

    """
    colors = list(map(lambda e: str(int(e)), y_true))

    if x_tensor.shape[1] < 2:
        raise AssertionError('x.shape[1] must be >= 2')

    elif x_tensor.shape[1] == 2:
        title = '<b>Initial Distribution</b>' if title is None else title
        fig = px.scatter(x=x_tensor[:, 0], y=x_tensor[:, 1], title=title, color=colors)

        if model is not None:
            dx = x_tensor[:, 0].max() - x_tensor[:, 0].min()
            dy = x_tensor[:, 1].max() - x_tensor[:, 1].min()

            x1 = torch.tensor([x_tensor[:, 0].min() - dx * k,
                               x_tensor[:, 1].min() - dy * k])

            x2 = torch.tensor([x_tensor[:, 0].max() + dx * k,
                               x_tensor[:, 1].max() + dy * k])

            flag, grid = _make_line_linear((x1[0] + dx * k / 2, x2[0] - dx * k / 2), model, threshold)

            if not flag:
                grid = _make_line(x1, x2, model, threshold, cnt_points, epsilon, insert_na)

            line_x, line_y = grid.detach().cpu().T

            fig.add_scatter(x=line_x, y=line_y, name='sep plane', mode='lines')

    else:
        title = '<b>TSNE of Distribution</b>' if title is None else title
        tsne_x = TSNE().fit_transform(x_tensor)
        fig = px.scatter(x=tsne_x[:, 0], y=tsne_x[:, 1], title=title, color=colors)

    fig.update_layout(**standard_layout)

    return fig


def _sort_points(line: torch.Tensor, epsilon: float = 1e-3, metric: int = 2, insert_na: bool = True) -> torch.Tensor:
    """
    Returns tensor sorted by closeness between each other. if || lines[i] - closest{lines[j]} ||_metric > epsilon
    insert [nan, nan]

    :param line: tensor n x 2
    :param epsilon: maximum closeness
    :param metric: l1, l2, or some other metric
    :param insert_na: na insertion flag
    :return: sorted tensor line with probably added nan values
    """

    copy_line = [line[0, :]]
    mask = torch.tile(torch.tensor([True]), line.shape[:1])
    mask[0] = False
    for i in range(line.shape[0] - 1):
        distances = torch.norm(line - copy_line[-1], p=metric, dim=1)
        distances[torch.logical_not(mask)] = torch.inf

        min_d, argmin_d = distances.min(), distances.argmin()
        if min_d <= epsilon ** 0.3 or insert_na is False:
            copy_line.append(line[[argmin_d]])
        else:
            copy_line.append(torch.tensor([torch.nan, torch.nan]))
            copy_line.append(line[[argmin_d]])

        mask[argmin_d] = False

    line = torch.zeros(len(copy_line), 2)
    for i in range(line.shape[0]):
        line[i, :] = copy_line[i]
    return line


roc_curve_plot = roc_curve_plot


def _make_line(x1: torch.Tensor, x2: torch.Tensor, model: torch.nn.Module, threshold: float = 0.5,
               cnt_points: int = 25, epsilon: float = 1e-3, insert_na: bool = True) -> torch.Tensor:
    """
    Returns x in [x1, x2] : threshold - epsilon <= model(x) <= threshold + epsilon

    :param x1: 2-dim tensor start
    :param x2: 2-dim tensor end
    :param model: some model that returns a torch tensor with class 1 probabilities using the call: model(x)
    :param threshold: if model(xi) >= threshold, then yi = 1
    :param cnt_points: number of points on each of the two axes
    :param epsilon: contour line points: :math:`\\{x\\in \\mathbb{R}^2 \\, | \\,
                \\text{threshold} - \\text{epsilon} \\le \\text{model}(x) \\le \\text{threshold} + \\text{epsilon}\\}`
    :param insert_na: na insertion flag
    :return: scatter plot go.Figure
    """
    if torch.isnan(x1[0]) or torch.isnan(x1[1]) or torch.isnan(x2[0]) or torch.isnan(x2[1]):
        return torch.tensor([[torch.nan, torch.nan]])

    lin_settings_1 = (min(x1[0], x2[0]), max(x1[0], x2[0]), cnt_points)
    lin_settings_2 = (min(x1[1], x2[1]), max(x1[1], x2[1]), cnt_points)

    grid = torch.cartesian_prod(torch.linspace(*lin_settings_1), torch.linspace(*lin_settings_2))

    with torch.no_grad():
        grid_pred = model(grid)

    mask = (threshold - epsilon <= grid_pred) & (grid_pred <= threshold + epsilon)
    if sum(mask) > 0:
        if sum(mask) > 1000:
            grid = grid[mask.flatten(), :]
            grid = grid[torch.linspace(0, grid.shape[0], 1000, dtype=torch.int64), :]
        else:
            grid = grid[mask.flatten(), :]
        grid = _sort_points(grid, epsilon=epsilon, insert_na=insert_na)
    else:
        grid = torch.tensor([torch.nan, torch.nan])
    return grid


def _make_line_linear(bounds_x: tuple[float, float],
                      model: torch.nn.Module,
                      threshold: float = 0.5) -> [bool, tuple | None]:
    """
    Returns for a linear model or a linear model with a sigmoid activation line on the plane

    :param bounds_x: bounds for x. tuple with two numbers
    :param model: linear model. e.g. SVM, Sigmoid
    :param threshold:
    :return:
    """
    # Check linear model
    if len(list(model.parameters())) > 2:
        return False, None

    try:
        w, b = model.parameters()
        w, b = w.flatten(), b.flatten()
    except Exception as e:
        print(e, 'non-linear model. is used basic _make_line')
        return False, None

    if len(w) > 2:
        return False, None

    x = torch.linspace(*bounds_x, 100)

    try:
        if hasattr(model, 'Sigmoid'):
            y = (torch.log(torch.tensor(threshold / (1 - threshold))) - b - w[0] * x) / w[1]

        else:
            y = (threshold - b - w[0] * x) / w[1]

    except Exception as e:
        print(e, 'non-linear model. is used basic _make_line')
        return False, None

    return True, torch.stack([x, y]).T



# Multi dimensional optimization algorithms for function optimization
from __future__ import annotations

from typing import Callable, Tuple, Sequence

import numpy
import torch

from .calculus import gradient, hessian, jacobian
from .one_optimize import brent
from .support import HistoryGD, update_history_gd, HiddenPrints, print_verbose


def initialize(function: Callable[[torch.Tensor], torch.Tensor],
               x0: torch.Tensor,
               epsilon: float = 1e-5,
               keep_history: bool = False) -> Tuple:
    """
    Returns initial x_k with double dtype, func_k, grad_k, round_precision, history

    :param function: callable that depends on the first positional argument
    :param x0: numpy ndarray which is initial approximation
    :param epsilon: optimization accuracy
    :param keep_history: flag of return history
    :return: tuple with x_k, func_k, grad_k, history, round_precision

    """
    x_k = x0.double().flatten()
    func_k = function(x_k)
    grad_k = gradient(function, x_k)
    round_precision = -int(numpy.log10(epsilon))  # variable to determine the rounding accuracy

    # if keep_history=True, we will save history. here is initial step
    if keep_history:
        history: HistoryGD = {
            'iteration': [0],
            'f_value': [func_k],
            'f_grad_norm': [grad_k.norm(2)],
            'x': [x_k.flatten()],
            'message': ''
        }

    else:
        history: HistoryGD = {'iteration': [], 'f_value': [], 'x': [], 'f_grad_norm': [], 'message': ''}

    return x_k, func_k, grad_k, history, round_precision


def bfgs(function: Callable[[torch.Tensor], torch.Tensor],
         x0: torch.Tensor,
         tolerance: float = 1e-8,
         max_iter: int = 500,
         verbose: bool = False,
         keep_history: bool = False) -> Tuple[torch.Tensor, HistoryGD]:
    """
    Returns a tensor n x 1 with optimal point and history using the BFGS method [1]_

    Broyden–Fletcher–Goldfarb–Shanno algorithm
    The algorithm does not use Wolfe conditions. Instead of wolfe, alg uses the optimal step.


    .. note::
        The algorithm only works for a flat x0, and the functions should depend on a flat array

    .. rubric:: References

    ..  [1] Wright and Nocedal, 'Numerical Optimization', 1999; pp.136-140 BFGS algorithm.

    :math:`\\rule{125mm}{0.2pt} \\\\`

    :param function: callable that depends on the first positional argument. Other arguments are passed through kwargs
    :param x0: start minimization point
    :param tolerance: criterion of stop os l2 norm(grad f) < tolerance
    :param max_iter: maximum number of iterations
    :param verbose: flag of printing iteration logs
    :param keep_history: flag of return history
    :return: tuple with point and history.

    .. rubric:: Examples

    .. code-block:: python3

        >>> def func(x): return 10 * x[0] ** 2 + x[1] ** 2 / 5
        >>> x_0 = torch.tensor([1, 2])
        >>> solution = bfgs(func, x_0)
        >>> print(solution[0])
        tensor([3.4372e-14, 1.8208e-14], dtype=torch.float64)
    """

    # initialization
    x_k, func_k, grad_k, history, round_precision = initialize(function, x0, tolerance, keep_history)
    h_k = torch.eye(x_k.shape[0], dtype=torch.float64) * tolerance ** 0.1
    grad_k = grad_k.reshape(-1, 1)
    x_k = x_k.reshape(-1, 1)
    s_k = x_k + 1

    # first verbose
    print_verbose(x_k, func_k, verbose, 0, round_precision)

    try:
        for i in range(max_iter):
            # stop criterion
            if grad_k.norm(2) < tolerance or s_k.norm(2) < tolerance:
                history['message'] = 'Searching finished. Successfully. code 0'
                return x_k.reshape(-1), history

            p_k = -h_k @ grad_k

            with HiddenPrints():
                def optimization_f(alpha: float | torch.Tensor) -> float | torch.Tensor:
                    return function(x_k + alpha * p_k)

                alpha_k = brent(optimization_f, (0, 1))[0]

            # step
            x_k_plus1 = x_k + alpha_k * p_k
            grad_f_k_plus1 = gradient(function, x_k_plus1, delta_x=1e-1).reshape(-1, 1)
            s_k = x_k_plus1 - x_k
            y_k = grad_f_k_plus1 - grad_k

            h_k = calc_h_new(h_k, s_k, y_k)
            grad_k = grad_f_k_plus1
            x_k = x_k_plus1
            func_k = function(x_k.flatten())

            # check divergence
            if torch.isnan(x_k).any():
                history['message'] = f'The method has diverged. code 2'
                return x_k.flatten(), history

            # verbose
            print_verbose(x_k.flatten(), func_k, verbose, i + 1, round_precision)

            # history
            if keep_history:
                history = update_history_gd(history, values=[i + 1, func_k, grad_k.norm(2), x_k.flatten()])
        else:
            history['message'] = 'Searching finished. Max iterations have been reached. code 1'

    except Exception as e:
        history['message'] = f'Optimization failed. {e}. code 2'

    return x_k.flatten(), history


def calc_h_new(h: torch.Tensor,
               s: torch.Tensor,
               y: torch.Tensor) -> torch.Tensor:
    """
    Calculates a new approximation of the inverse Hessian matrix

    :param h: The previous approximation of the H matrix
    :param s: the difference x_{i+1} - x_{i}
    :param y: the difference f'_{i+1} - f'_{i}
    :return: The new approximation of inverse Hessian matrix
    """

    ro = 1 / (y.T @ s)
    i = torch.eye(h.shape[0]).double()
    h_new = (i - ro * s @ y.T) @ h @ (i - ro * s @ y.T) + ro * s @ s.T

    return h_new


def gd_constant(function: Callable[[torch.Tensor], torch.Tensor],
                x0: torch.Tensor,
                epsilon: float = 1e-5,
                gamma: float = 0.1,
                max_iter: int = 500,
                verbose: bool = False,
                keep_history: bool = False) -> Tuple[torch.Tensor, HistoryGD]:
    """
    Returns a tensor n x 1 with optimal point and history using
    Algorithm with constant step.
    The gradient of the function shows us the direction of increasing the function.
    The idea is to move in the opposite direction to :math:`\\displaystyle x_{k + 1} \\text{ where }
    f(x_{k + 1}) < f(x_{k}) \\text{ .}`

    But, if we add a gradient to :math:`\\displaystyle x_{k}` without changes, our method will often diverge.
    So we need to add a gradient with some weight :math:`\\displaystyle \\gamma \\text{ .}\\\\`


    :param function: callable that depends on the first positional argument
    :param x0: Torch tensor which is initial approximation
    :param epsilon: optimization accuracy
    :param gamma: gradient step
    :param max_iter: maximum number of iterations
    :param verbose: flag of printing iteration logs
    :param keep_history: flag of return history
    :return: tuple with point and history.

    .. rubric:: Examples

    .. code-block:: python3

        >>> def func(x): return x[0] ** 2 + x[1] ** 2
        >>> x_0 = torch.tensor([1, 2])
        >>> solution = gd_constant(func, x_0)
        >>> print(solution[0])
        tensor([1.9156e-06, 3.8312e-06], dtype=torch.float64)
    """

    # initialization
    x_k, func_k, grad_k, history, round_precision = initialize(function, x0, epsilon, keep_history)

    # first verbose
    print_verbose(x_k, func_k, verbose, 0, round_precision)

    try:
        for i in range(max_iter - 1):

            if grad_k.norm(2) < epsilon:  # comparing of norm 2 with optimization accuracy
                history['message'] = 'Optimization terminated successfully. code 0'
                break
            else:
                x_k = x_k - gamma * grad_k  # updating the point for next iter and repeat
                grad_k = gradient(function, x_k)
                func_k = function(x_k)

            # verbose
            print_verbose(x_k, func_k, verbose, i + 1, round_precision)

            # history
            if keep_history:
                history = update_history_gd(history, values=[i + 1, func_k, (grad_k ** 2).sum() ** 0.5, x_k])

        else:
            history['message'] = 'Optimization terminated. Max steps. code 1'

    except Exception as e:
        history['message'] = f'Optimization failed. {e}. code 2'

    return x_k, history


def gd_frac(function: Callable[[torch.Tensor], torch.Tensor],
            x0: torch.Tensor,
            epsilon: float = 1e-5,
            gamma: float = 0.1,
            delta: float = 0.1,
            lambda0: float = 0.1,
            max_iter: int = 500,
            verbose: bool = False,
            keep_history: bool = False) -> Tuple[torch.Tensor, HistoryGD]:
    """
    Returns a tensor n x 1 with optimal point and history using
    Algorithm with fractional step.

    Requirements: :math:`\\ 0 < \\lambda_0 < 1`  is the step multiplier, :math:`0 < \\delta < 1` influence on step size.

    :param function: callable that depends on the first positional argument
    :param x0: Torch tensor which is initial approximation
    :param epsilon: optimization accuracy
    :param gamma: gradient step
    :param delta: value of the crushing parameter
    :param lambda0: initial step
    :param max_iter: maximum number of iterations
    :param verbose: flag of printing iteration logs
    :param keep_history: flag of return history
    :return: tuple with point and history.



    .. rubric:: Examples

    .. code-block:: python3

        >>> def func(x): return x[0] ** 2 + x[1] ** 2
        >>> x_0 = torch.tensor([1, 2])
        >>> solution = gd_frac(func, x_0)
        >>> print(solution[0])
        tensor([1.9156e-06, 3.8312e-06], dtype=torch.float64)

    """

    # initialization
    x_k, func_k, grad_k, history, round_precision = initialize(function, x0, epsilon, keep_history)

    # first verbose
    print_verbose(x_k, func_k, verbose, 0, round_precision)

    try:
        for i in range(max_iter - 1):

            # point for first comparison, first gradient step
            t = x_k - gamma * grad_k
            func_t = function(t)

            # will divide the gradient step until condition is met
            while not func_t - func_k <= - gamma * delta * sum(grad_k ** 2):
                gamma = gamma * lambda0
                t = x_k - gamma * grad_k
                func_t = function(t)

            x_k = t
            func_k = func_t
            grad_k = gradient(function, x_k)

            #  comparing of norm 2 with optimization accuracy
            if grad_k.norm(2) < epsilon:
                history['message'] = 'Optimization terminated successfully. code 0'
                break

            # verbose
            print_verbose(x_k, func_k, verbose, i + 1, round_precision)

            # history
            if keep_history:
                history = update_history_gd(history, values=[i + 1, func_k, (grad_k ** 2).sum() ** 0.5, x_k])

        else:
            history['message'] = 'Optimization terminated. Max steps. code 1'

    except Exception as e:
        history['message'] = f'Optimization failed. {e}. code 2'

    return x_k, history


def gd_optimal(function: Callable[[torch.Tensor], torch.Tensor],
               x0: torch.Tensor,
               epsilon: float = 1e-5,
               max_iter: int = 500,
               verbose: bool = False,
               keep_history: bool = False) -> Tuple[torch.Tensor, HistoryGD]:
    """
    Returns a tensor n x 1 with optimal point and history using
    Algorithm with optimal step.
    The idea is to choose a gamma that minimizes the function in the direction :math:`\\ f'(x_k)`



    :param function: callable that depends on the first positional argument
    :param x0: Torch tensor which is initial approximation
    :param epsilon: optimization accuracy
    :param max_iter: maximum number of iterations
    :param verbose: flag of printing iteration logs
    :param keep_history: flag of return history
    :return: tuple with point and history.

    .. rubric:: Examples

    .. code-block:: python3

        >>> def func(x): return -torch.exp(- x[0] ** 2 - x[1] ** 2)
        >>> x_0 = torch.tensor([1, 2])
        >>> solution = gd_optimal(func, x_0)
        >>> print(solution[0])
        tensor([9.2070e-08, 1.8405e-07], dtype=torch.float64)
    """

    # initialization
    x_k, func_k, grad_k, history, round_precision = initialize(function, x0, epsilon, keep_history)

    # first verbose
    print_verbose(x_k, func_k, verbose, 0, round_precision)

    try:
        for i in range(max_iter - 1):

            # comparing of norm 2 with optimization accuracy
            if grad_k.norm(2) < epsilon:
                history['message'] = 'Optimization terminated successfully. code 0'
                break

            with HiddenPrints():  # hiding the prints of the results of the brent algorithm
                def optimization_f(gam: float | torch.Tensor) -> float | torch.Tensor:
                    return function(x_k - gam * grad_k)

                gamma = brent(optimization_f, (0, 1))[0]

            x_k = x_k - gamma * grad_k
            grad_k = gradient(function, x_k)
            func_k = function(x_k)

            # verbose
            print_verbose(x_k, func_k, verbose, i + 1, round_precision)

            # history
            if keep_history:
                history = update_history_gd(history, values=[i + 1, func_k, (grad_k ** 2).sum() ** 0.5, x_k])

        else:
            history['message'] = 'Optimization terminated. Max steps. code 1'

    except Exception as e:
        history['message'] = f'Optimization failed. {e}. code 2'

    return x_k, history


def nonlinear_cgm(function: Callable[[torch.Tensor], torch.Tensor],
                  x0: torch.Tensor,
                  epsilon: float = 1e-5,
                  max_iter: int = 500,
                  verbose: bool = False,
                  keep_history: bool = False) -> Tuple[torch.Tensor, HistoryGD]:
    """
    Returns a tensor n x 1 with optimal point and history.
    Algorithm works when the function is approximately quadratic near the minimum, which is the case when the
    function is twice differentiable at the minimum and the second derivative is non-singular there [1]_

    .. rubric:: References

    ..  [1] Nocedal, J., &amp; Wright, S. J. (2006). 5.2 NONLINEAR CONJUGATE GRADIENT METHOD
        In Numerical optimization (pp. 121). essay, Springer.

    :param function: callable that depends on the first positional argument
    :param x0: Torch tensor which is initial approximation
    :param epsilon: optimization accuracy
    :param max_iter: maximum number of iterations
    :param verbose: flag of printing iteration logs
    :param keep_history: flag of return history
    :return: tuple with point and history.


    .. rubric:: Examples

    Example for :math:`f(x, y) = (x + 0.5)^2 + (y - 0.5)^2, \\quad x = 1`

    .. code-block:: python3

        >>> def func(x): return 10 * x[0] ** 2 + x[1] ** 2 / 5
        >>> x_0 = torch.tensor([1, 2])
        >>> solution = nonlinear_cgm(func, x_0)
        >>> print(solution[0])
        tensor([6.9846e+25, 4.2454e+26], dtype=torch.float64)
    """

    # initialization
    x_k, func_k, grad_k, history, round_precision = initialize(function, x0, epsilon, keep_history)
    p_k = grad_k

    # first verbose
    print_verbose(x_k, func_k, verbose, 0, round_precision)

    try:
        for i in range(max_iter - 1):

            if grad_k.norm(2) < epsilon:
                history['message'] = 'Optimization terminated successfully. code 0'
                break
            else:
                with HiddenPrints():
                    def optimization_f(gam: float | torch.Tensor) -> float | torch.Tensor:
                        return function(x_k - gam * grad_k)

                    gamma = brent(optimization_f, (0, 1))[0]

                x_k = x_k - gamma * p_k
                grad_k_new = gradient(function, x_k)
                beta_fr = (grad_k_new @ grad_k_new.reshape(-1, 1)) / (grad_k @ grad_k.reshape(-1, 1))
                p_k = grad_k_new + beta_fr * p_k
                grad_k = grad_k_new
                func_k = function(x_k)

                # verbose
                print_verbose(x_k, func_k, verbose, i + 1, round_precision)

                # history
                if keep_history:
                    history = update_history_gd(history, values=[i + 1, func_k, grad_k.norm(2), x_k])

        else:
            history['message'] = 'Optimization terminated. Max steps. code 1'

    except Exception as e:
        history['message'] = f'Optimization failed. {e}. code 2'

    return x_k, history


def log_barrier_function(function: Callable[[torch.Tensor], torch.Tensor],
                         x0: torch.Tensor,
                         mu: float,
                         inequality_constraints: Sequence[Callable[[torch.Tensor], torch.Tensor]]
                         ) -> torch.Tensor:
    """
    Support function. Compute log-barrier function

    .. math::

        P(x, \\mu) = f(x) - \\mu \\sum_{i\\in\\mathcal{I}}\\ln c_i(x)

    :param function: callable that depends on the first positional argument
    :param x0: some specific point x(Torch tensor)
    :param mu: parameter weighing constraints
    :param inequality_constraints: :math:`\\mathcal{I}` is set of inequality functions
    :return: tuple with point and history.
    """

    m = len(inequality_constraints)  # Amount of inequality constraints
    output_lb = function(x0)
    for j in range(m):
        const_function = inequality_constraints[j](x0)

        if 0 <= const_function < 1e-8:
            output_lb += mu * 0
        elif const_function > 1e-8:
            output_lb -= mu * torch.log(const_function)
        else:
            output_lb += 10 ** 10

    return output_lb


def primal_dual_interior(function: Callable[[torch.Tensor], torch.Tensor],
                         x0: torch.Tensor,
                         inequality_constraints: Sequence[Callable[[torch.Tensor], torch.Tensor]],
                         mu: float = 1e-4,
                         epsilon: float = 1e-12,
                         alpha: float = 1e-1,
                         max_iter: int = 200,
                         verbose: bool = False,
                         keep_history: bool = False) -> Tuple[torch.Tensor, HistoryGD]:
    """
    Returns point and history of minimization. [1]_

    AIM: minimize :math:`\\ f(x)` subject to :math:`\\ c(x) \\geqslant 0`; c from inequality_constraints

    :math:`\\ B(x,\\mu) = f(x) - \\mu \\sum_{i=1}^m \\log(c_i(x)) \\rightarrow \\min`

    Here :math:`\\mu` is a small positive scalar, :math:`\\mu \\rightarrow 0`

    :param function: callable that depends on the first positional argument
    :param x0: some specific point x(Torch tensor)
    :param inequality_constraints: :math:`\\mathcal{I}` is set of inequality functions
    :param mu: is a small positive scalar, sometimes called the "barrier parameter"
    :param epsilon: optimization accuracy
    :param alpha: step length
    :param max_iter: maximum number of iterations
    :param verbose: flag of printing iteration logs
    :param keep_history: flag of return history
    :return:  tuple with point and history.

    :raises ArithmeticError: if x0 is not in trust region.

    .. rubric:: Reference

    .. [1] https://en.wikipedia.org/wiki/Interior-point_method

    .. rubric:: Examples

    .. code-block:: python3


        >>> primal_dual_interior(lambda x: (x[0] + 0.5) ** 2 + (x[1] - 0.5) ** 2, torch.tensor([0.9, 0.1]),
        >>>                    [lambda x: x[0], lambda x: 1 - x[0], lambda x: x[1], lambda x: 1 - x[1]])[0]
        tensor([1.9910e-04, 5.0000e-01], dtype=torch.float64)
    """

    # initialization
    def lb_function(x: torch.Tensor, _mu: float = mu) -> torch.Tensor | float:
        """ log barrier function -- main function. therefore, instead of a function, we use lb_function """
        return log_barrier_function(function, x, _mu, inequality_constraints)

    x_k, func_k, grad_k, history, round_precision = initialize(function, x0, epsilon, keep_history)
    m = len(inequality_constraints)  # Amount of inequality constraints
    n = x_k.shape[0]  # Amount of variables
    p = torch.ones(m + n)  # init of p
    lambdas = torch.rand(m, dtype=torch.float64)

    try:
        function(x0)
        for i in range(m):
            if inequality_constraints[i](x0) < 0:
                raise ArithmeticError

    except ArithmeticError:
        history['message'] = 'Point out of domain'
        return x_k, history

    # first verbose
    print_verbose(x_k, func_k, verbose, 0, round_precision)
    try:
        for i in range(max_iter):

            # terminate condition
            if p[:n].norm(2) < epsilon:
                history['message'] = 'Optimization terminated successfully. code 0'
                break

            w = hessian(lb_function, x_k, 1e-4)  # hessian of lb function
            a = jacobian(inequality_constraints, x_k)
            c = torch.diag(torch.tensor([c(x_k) for c in inequality_constraints])).double()

            left_matrix = torch.zeros(n + m, n + m, dtype=torch.float64)  # left matrix equation
            left_matrix[:n, :n] = w
            left_matrix[:n, n:] = -a.T
            left_matrix[n:, :n] = torch.diag(lambdas) @ a
            left_matrix[n:, n:] = c

            right_matrix = torch.zeros(n + m, 1, dtype=torch.float64)  # right matrix
            # - g + A.T @ lambdas
            right_matrix[:n] = - gradient(lb_function, x_k).reshape(-1, 1) + a.T @ lambdas.reshape(-1, 1)
            # mu 1 - c @ lambdas
            right_matrix[n:] = mu * torch.ones(m, 1) - c @ lambdas.reshape(-1, 1)

            try:  # check singularity
                p = torch.linalg.solve(left_matrix, right_matrix.flatten())  # direction

            except torch.linalg.LinAlgError:
                left_matrix[:n, :n] += 1e-2 * torch.eye(n)
                left_matrix[n:, n:] += 1e-4 * torch.eye(m)

                p = torch.linalg.solve(left_matrix, right_matrix.flatten())  # direction

            x_k += alpha * p[:n]
            lambdas += alpha * p[n:]

            # verbose
            if verbose or keep_history:
                func_k = function(x_k)

            print_verbose(x_k, func_k, verbose, i + 1, round_precision)

            # history
            if keep_history:
                history = update_history_gd(history, values=[i + 1, func_k, p[:n].norm(2), x_k.clone()])

            mu *= 0.99

        else:
            history['message'] = 'Optimization terminated. Max steps. code 1'

    except Exception as e:
        history['message'] = f'Optimization failed. {e}. code 2'

    return x_k, history


def log_barrier_solver(function: Callable[[torch.Tensor], torch.Tensor],
                       x0: torch.Tensor,
                       inequality_constraints: Sequence[Callable[[torch.Tensor], torch.Tensor]],
                       epsilon: float = 1e-5,
                       max_iter: int = 1000,
                       keep_history: bool = False,
                       verbose: bool = False) -> Tuple[torch.Tensor, HistoryGD]:
    """
    Returns optimal point of optimization with inequality constraints by Log Barrier method [1]_




    .. rubric:: References

    ..  [1] Nocedal, J., &amp; Wright, S. J. (2006). 19.6 THE PRIMAL LOG-BARRIER METHOD.
            In Numerical optimization (pp. 583–584). essay, Springer.

    :math:`\\rule{125mm}{0.2pt} \\\\`

    :param function: callable that depends on the first positional argument
    :param x0: some specific point x(Torch tensor)
    :param epsilon: optimization accuracy
    :param inequality_constraints: :math:`\\mathcal{I}` is set of inequality functions
    :param max_iter: maximum number of iterations
    :param keep_history: flag of return history
    :param verbose: flag of printing iteration logs
    :return: tuple with point and history.


    .. rubric:: Examples

    Example for :math:`f(x, y) = (x + 0.5)^2 + (y - 0.5)^2, \\quad 0 \\le x \\le 1, 0 \\le y \\le 1`

    .. code-block:: python3

        >>> log_barrier_solver(lambda x: (x[0] + 0.5) ** 2 + (x[1] - 0.5) ** 2, torch.tensor([0.9, 0.1]),
        >>>                    [lambda x: x[0], lambda x: 1 - x[0], lambda x: x[1], lambda x: 1 - x[1]])
        tensor([0.0032, 0.5000], dtype=torch.float64)

    """
    m = len(inequality_constraints)  # Amount of inequality constraints
    x_k, func_k, grad_k, history, round_precision = initialize(function, x0, epsilon, keep_history)

    try:
        function(x0)
        for i in range(m):
            if inequality_constraints[i](x0) < 0:
                raise ArithmeticError

    except ArithmeticError:
        history['message'] = 'Point out of domain'
        return x_k, history

    tau = 1  # The tau sequence will be geometric

    try:
        for i in range(max_iter):
            mu_k = tau ** 0.5
            x_k, history_step = gd_frac(lambda x: log_barrier_function(function, x, mu_k, inequality_constraints),
                                        x_k, gamma=mu_k, epsilon=tau, keep_history=True, max_iter=5)

            tau *= 0.9
            if tau <= epsilon:
                history['message'] = 'Optimization terminated successfully. code 0'
                break

            # verbose
            if verbose or keep_history:
                func_k = function(x_k)
                grad_k = gradient(function, x_k)

            print_verbose(x_k, func_k, verbose, i + 1, round_precision)

            # history
            if keep_history:
                history = update_history_gd(history, values=[i + 1, func_k, grad_k.norm(2), x_k.clone()])

    except Exception as e:
        history['message'] = f'Optimization failed. {e}. code 2'

    return x_k, history


def constrained_lagrangian_solver(function: Callable[[float | torch.Tensor], torch.Tensor],
                                  x0: torch.Tensor,
                                  constraints: Sequence[Callable[[float | torch.Tensor], torch.Tensor]],
                                  x_bounds: Sequence[Tuple[float, float]] | None | torch.Tensor = None,
                                  epsilon: float = 1e-4,
                                  max_iter: int = 250,
                                  keep_history: bool = False,
                                  verbose: bool = False) -> Tuple[torch.Tensor, HistoryGD]:
    """
    Returns a tensor n x 1 with optimal point and history of minimization by newton_eq_const.
    Alias of ''Newton’s method under equality constrains'' [1]_

    Example for :math:`f(x, y) = (x + 0.5)^2 + (y - 0.5)^2, \\quad x = 1`

    .. rubric:: References

    ..  [1] Nocedal, J., &amp; Wright, S. J. (2006). 17.4 PRACTICAL AUGMENTED LAGRANGIAN METHODS.
        In Numerical optimization (pp. 519–521). essay, Springer.

    :math:`\\rule{125mm}{0.2pt} \\\\`

    :param function: callable that depends on the first positional argument
    :param x0: some specific point x(Torch tensor)
    :param constraints: list of equality constraints
    :param x_bounds: bounds on x. e.g. 0 <= x[i] <= 1, then x_bounds[i] = (0, 1)
    :param epsilon: optimization accuracy
    :param max_iter: maximum number of iterations
    :param keep_history: flag of return history
    :param verbose: flag of printing iteration logs
    :return: tuple with point and history.

    .. rubric:: Examples

    .. code-block:: python3

        >>> constrained_lagrangian_solver(lambda x: (x[0] + 0.5) ** 2 + (x[1] - 0.5) ** 2,
        >>>                              torch.tensor([0.1, 0.1]),[lambda x: x[0] - 1]))
        tensor([1.0540, 0.5000], dtype=torch.float64)
    """
    m = len(constraints)
    if x_bounds is None:
        x_bounds = torch.zeros(len(x0), 2).double()
        for i in range(len(x0)):
            x_bounds[i, :] = torch.tensor([-torch.inf, torch.inf])

    elif isinstance(x0, torch.Tensor):
        assert x0.shape[0] == x_bounds.shape[0], 'the boundaries should be n x 2 tensor'
        assert len(x_bounds.shape) == 2, 'the boundaries should be n x 2 tensor'
        assert x_bounds.shape[1] == 2, 'the boundaries should be n x 2 tensor'

    else:
        assert len(x_bounds) == x0.shape[0], 'the boundaries should be for each variable'

    if not isinstance(x0, torch.Tensor):

        _x_bounds = torch.zeros(x0.shape[0], 2).double()
        for i, const in enumerate(x_bounds):
            _x_bounds[i, :] = const
        x_bounds = _x_bounds

    def c(x: torch.Tensor) -> torch.Tensor:
        """Returns vector of constraints at specific x"""
        _c = torch.zeros(m).double()
        for j in range(m):
            _c[j] = constraints[j](x)
        return _c

    def lagrangian_a(x: torch.Tensor, lam: torch.Tensor, mu: float) -> torch.Tensor:
        """
        Returns :math:`\\mathcal{L}_a`
        Nocedal, J., &amp; Wright, S. J. (2006). Numerical optimization (p. 520)
        """

        output = function(x)

        for j in range(m):
            output += -lam[j] * constraints[j](x) + mu / 2 * constraints[j](x) ** 2

        return output

    def p_function(g: torch.Tensor, u_l_bounds: torch.Tensor):
        """
        P(g, l, u) is the projection of the vector g :math:`\\in` IRn onto the rectangular box :math:`[l, u]`
        Nocedal, J., &amp; Wright, S. J. (2006). Numerical optimization (p. 520)
        """

        # for j in range(len(g)):
        #     if g[j] >= u_l_bounds[j][1]:
        #         g[j] = u_l_bounds[j][1]
        #     elif g[j] <= u_l_bounds[j][0]:
        #         g[j] = u_l_bounds[j][0]

        # replaced by

        g = torch.minimum(torch.maximum(g, u_l_bounds[:, 0]), u_l_bounds[:, 1])

        return g

    x_k, func_k, grad_k, history, round_precision = initialize(function, x0, epsilon, keep_history)

    try:
        function(x0)
        for i in range(m):
            constraints[i](x0)

    except ArithmeticError:
        history['message'] = 'Point out of domain'
        return x_k, history

    lambdas_k = torch.rand(m)
    eta = epsilon  # Main tolerance for constraints
    omega = eta  # Main tolerance for lagrange function
    mu_k = 10
    omega_k = 1 / mu_k
    eta_k = 1 / mu_k ** 0.1

    for i in range(max_iter):

        def local_min_function(x):
            grad_lagrangian = x - gradient(lambda y: lagrangian_a(y, lambdas_k, mu_k), x)
            p = p_function(grad_lagrangian, x_bounds)
            return (x - p).norm(2)

        x_k = nonlinear_cgm(local_min_function, x_k, epsilon=omega_k, max_iter=4)[0]

        if verbose or keep_history:
            func_k = function(x_k)
            grad_k = gradient(function, x_k)

        print_verbose(x_k, func_k, verbose, i + 1, round_precision)

        # history
        if keep_history:
            history = update_history_gd(history, values=[i + 1, func_k, grad_k.norm(2), x_k.clone()])

        c_k = c(x_k)
        if c_k.norm(2) <= eta_k:
            # test for convergence
            if c_k.norm(2) <= eta and local_min_function(x_k) <= omega:
                break

            # update multipliers, tighten tolerances
            lambdas_k = lambdas_k - mu_k * c_k

        else:
            # increase penalty parameter, tighten tolerances
            lambdas_k = lambdas_k - mu_k * c_k
            mu_k = mu_k * 100
            eta_k = 1 / mu_k ** 0.1
            omega = 1 / mu_k

    return x_k, history



from .optimize import NueSGD, SimulatedAnnealing
from .classification import LogisticRegressionRBF, LogisticRegression, SVM
from . import metrics
from .regression import LinearRegression, ExponentialRegression, PolynomialRegression, PolyTransform

__all__ = [
    "NueSGD",
    "SimulatedAnnealing",
    "LogisticRegression",
    "LogisticRegressionRBF",
    "LinearRegression",
    "ExponentialRegression",
    "PolyTransform",
    "PolynomialRegression",
    "SVM"
]



from __future__ import annotations

import os
import sys
from typing import Text, Sequence, Any

if sys.version_info >= (3, 8):
    from typing import TypedDict, Literal  # pylint: disable=no-name-in-module
else:
    from typing_extensions import TypedDict, Literal

import torch


class HistoryGSS(TypedDict):
    """
    Class with an optimization history of Golden section search
    """
    iteration: list
    middle_point: list[float]
    f_value: list[float]
    left_point: list[float]
    right_point: list[float]


class HistorySPI(TypedDict):
    """
    Class with an optimization history of Successive parabolic interpolation
    """
    iteration: list[int]
    f_value: list[float]
    x0: list[float]
    x1: list[float]
    x2: list[float]


class HistoryBrent(TypedDict):
    """
    Class with an optimization history of Brant's algorithm
    """
    iteration: list[int]

    f_least: list[float]
    f_middle: list[float]
    f_largest: list[float]

    x_least: list[float]
    x_middle: list[float]
    x_largest: list[float]

    left_bound: list[float]
    right_bound: list[float]
    type_step: list[Text]


def update_history_brent(history: HistoryBrent, values: Sequence[Any]) -> HistoryBrent:
    """
    Updates brent history
    :param history: HistoryBrent object in which the update is required
    :param values: Sequence with values: 'iteration', 'f_least', 'f_middle', 'f_largest',  'x_least',
                                         'x_middle', 'x_largest', 'left_bound', 'right_bound', 'type_step'
    :return: updated HistoryBrent
    """
    name: Literal['iteration', 'f_least', 'f_middle', 'f_largest', 'x_least', 'x_middle',
                  'x_largest', 'left_bound', 'right_bound', 'type_step']

    for i, name in enumerate(['iteration', 'f_least', 'f_middle', 'f_largest', 'x_least', 'x_middle',
                              'x_largest', 'left_bound', 'right_bound', 'type_step']):
        history[name].append(values[i])

    return history


class HiddenPrints:
    """
    Object hides print. Working with context manager "with"::

        >>> with HiddenPrints():
        >>>     print("It won't be printed")
    """

    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


class HistoryGD(TypedDict):
    """
    Class with an optimization history of gradient descent methods
    """
    iteration: list
    f_value: list
    f_grad_norm: list
    x: list[torch.Tensor]
    message: Text


def update_history_gd(history: HistoryGD, values: list) -> HistoryGD:
    """
    Update HistoryMDO with values, which contains iteration, f_value, f_grad_norm, x as a list

    :param history: object of HistoryGD
    :param values: new values that need to append in history in order iteration, f_value, f_grad_norm, x
    :return: updated history
    """
    key: Literal['iteration', 'f_value', 'f_grad_norm', 'x']
    for i, key in enumerate(['iteration', 'f_value', 'f_grad_norm', 'x']):
        history[key].append(values[i])
    return history


def print_verbose(x_k: torch.Tensor,
                  func_k: torch.Tensor | float,
                  verbose: bool,
                  iteration: int,
                  round_precision: int) -> None:
    """
    Prints iteration verbose

    :param x_k: specific point
    :param func_k: float or 0-d tensor
    :param verbose: flag of print verbose
    :param iteration: number of iteration
    :param round_precision: precision of printing float numbers
    :return: none
    """
    if verbose:
        print(f'iteration: {iteration:4d}  |  '
              f'x = [{", ".join(map(lambda x: f"{round(float(x), 4):>10.4f}", x_k))}]  |  '
              f'f(x) = {round(float(func_k), round_precision)}')



from __future__ import annotations

from typing import Callable

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import torch

from ..multi_optimize import gd_optimal, bfgs
from ..support import HistoryGD


def make_contour(function: Callable[[torch.Tensor], float | torch.Tensor],
                 bounds: tuple[tuple[float, float], tuple[float, float]],
                 cnt_dots: int = 100,
                 colorscale='teal',
                 showlegend: bool = False) -> go.Contour:
    """
    Return go.Contour for draw by go.Figure. Evaluate function per each point in the 2d grid

    :param function: callable that depends on the first positional argument
    :param bounds: two tuples with constraints for x- and y-axis
    :param cnt_dots: number of point per each axis
    :param colorscale: plotly colorscale for go.Contour
    :param showlegend: show legend flag
    :return: go.Contour

    .. code-block:: python3

        >>> def f(x): return x[0] ** 2 + x[1] ** 2 / 2
        >>> make_contour(f, ((0, 1), (0, 1)), cnt_dots=4)

    .. code-block:: python3

        Contour({
            'colorscale': [[0.0, 'rgb(209, 238, 234)'], [0.16666666666666666, 'rgb(168,
                           219, 217)'], [0.3333333333333333, 'rgb(133, 196, 201)'], [0.5,
                           'rgb(104, 171, 184)'], [0.6666666666666666, 'rgb(79, 144,
                           166)'], [0.8333333333333334, 'rgb(59, 115, 143)'], [1.0,
                           'rgb(42, 86, 116)']],
            'name': 'f(x, y)',
            'showlegend': False,
            'showscale': False,
            'x': array([0.        , 0.33333334, 0.6666666 , 1.        ], dtype=float32),
            'y': array([0.        , 0.33333334, 0.6666666 , 1.        ], dtype=float32),
            'z': array([[0.        , 0.11111112, 0.4444444 , 1.        ],
                        [0.05555556, 0.16666669, 0.49999994, 1.0555556 ],
                        [0.2222222 , 0.3333333 , 0.66666657, 1.2222222 ],
                        [0.5       , 0.6111111 , 0.9444444 , 1.5       ]], dtype=float32)
        })

    """
    x_axis, y_axis, z_axis = _make_axis(function, bounds, cnt_dots)

    return go.Contour(x=x_axis, y=y_axis, z=z_axis, colorscale=colorscale,
                      name='f(x, y)', showscale=showlegend, showlegend=showlegend)


def gen_simple_gradient(function: Callable[[torch.Tensor], torch.Tensor],
                        history: HistoryGD,
                        cnt_dots: int = 200,
                        title: str = '<b>Contour plot with optimization steps</b>',
                        showlegend: bool = True,
                        font_size: int = 18) -> go.Figure:
    """
    Return go.Figure with gradient steps under contour plot. Not animated

    :param function: callable that depends on the first positional argument
    :param history: History after some gradient method
    :param cnt_dots: the numbers of point per each axis
    :param title: title of chart
    :param showlegend: flag of showing legend
    :param font_size: font size
    :return: go.Figure with contour and line of gradient steps

    .. code-block:: python3

        >>> def f(x): return x[0] ** 2 + x[1] ** 2 / 2
        >>> x_opt, hist = gd_optimal(f, torch.tensor([8, 5]), keep_history=True)
        >>> gen_simple_gradient(f, hist).show()

    """

    descent_history = _make_descent_history(history)
    bounds = _make_ranges(history)

    layout = go.Layout(
        title=title,
        xaxis={'title': r'<b>x</b>', 'color': 'black'},
        yaxis={'title': r'<b>y</b>', 'color': 'black'},
        font={'size': font_size, 'color': 'black'},
        legend={'font': {'color': 'black'}}
    )

    contour = make_contour(function=function, bounds=bounds, cnt_dots=cnt_dots, showlegend=False)

    descending_way = go.Scatter(
        x=descent_history.x,
        y=descent_history.y,
        name='descent',
        mode='lines+markers',
        line={'width': 3, 'color': 'rgb(202, 40, 22)'},
        marker={'size': 10, 'color': 'rgb(202, 40, 22)'},
        showlegend=showlegend
    )

    fig = go.Figure(data=[contour, descending_way], layout=layout)

    return fig


def make_surface(function: Callable[[torch.Tensor], float | torch.Tensor],
                 bounds: tuple[tuple[float, float], tuple[float, float]],
                 cnt_dots: int = 100,
                 colorscale='teal',
                 showlegend: bool = False) -> go.Surface:
    """
    Return go.Surface for draw by go.Figure. Evaluate function per each point in the 2d grid

    :param function: callable that depends on the first positional argument
    :param bounds: two tuples with constraints for x- and y-axis
    :param cnt_dots: number of point per each axis
    :param colorscale: plotly colorscale for go.Contour
    :param showlegend: showlegend flag
    :return: go.Surface

    .. code-block:: python3

        >>> def f(x): return x[0] ** 2 + x[1] ** 2 / 2
        >>> make_surface(f, ((0, 1), (0, 1)), cnt_dots=4)

    .. code-block:: python3

        Surface({
            'colorscale': [[0.0, 'rgb(209, 238, 234)'], [0.16666666666666666, 'rgb(168,
                           219, 217)'], [0.3333333333333333, 'rgb(133, 196, 201)'], [0.5,
                           'rgb(104, 171, 184)'], [0.6666666666666666, 'rgb(79, 144,
                           166)'], [0.8333333333333334, 'rgb(59, 115, 143)'], [1.0,
                           'rgb(42, 86, 116)']],
            'name': 'f(x, y)',
            'opacity': 0.75,
            'showlegend': False,
            'x': array([0.        , 0.33333334, 0.6666666 , 1.        ], dtype=float32),
            'y': array([0.        , 0.33333334, 0.6666666 , 1.        ], dtype=float32),
            'z': array([[0.        , 0.11111112, 0.4444444 , 1.        ],
                        [0.05555556, 0.16666669, 0.49999994, 1.0555556 ],
                        [0.2222222 , 0.3333333 , 0.66666657, 1.2222222 ],
                        [0.5       , 0.6111111 , 0.9444444 , 1.5       ]], dtype=float32)
        })

    """

    x_axis, y_axis, z_axis = _make_axis(function, bounds, cnt_dots)

    return go.Surface(x=x_axis, y=y_axis, z=z_axis, colorscale=colorscale,
                      name='f(x, y)', opacity=0.75, showlegend=showlegend)


def _check_bounds(bounds: tuple[tuple[float, float], tuple[float, float]]):
    """
    Assert if bounds not have 2 tuple. In 2 tuples exists 2 floats

    :param bounds: two tuples with constraints for x- and y-axis
    """
    assert len(bounds) == 2, 'two tuples are required'
    assert len(bounds[0]) == 2 and len(bounds[1]) == 2, 'both tuples must have 2 numbers'


def _make_z(function: Callable[[torch.Tensor], float],
            x_axis: torch.Tensor,
            y_axis: torch.Tensor,
            cnt_dots) -> torch.Tensor:
    """
    Returns matrix, with calculated function at each point from grid square bounds[0] x bounds[1]

    :param function: callable that depends on the first positional argument
    :param x_axis: values for x-axis
    :param y_axis: values for y-axis
    :param cnt_dots: number of point per each axis
    :return: tensor with function values
    """

    z_axis = torch.zeros(cnt_dots, cnt_dots).float()
    for i, y in enumerate(y_axis):
        z_axis_i = []
        for x in x_axis:
            z_axis_i.append(function(torch.tensor([x, y])))

        z_axis[i, :] = torch.tensor(z_axis_i).float()

    return z_axis


def _make_axis(function: Callable[[torch.Tensor], float],
               bounds: tuple[tuple[float, float], tuple[float, float]],
               cnt_dots: int) -> [torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Returns x, y, z transformed

    :param function: callable that depends on the first positional argument
    :param bounds: constraints for each axis

    :param cnt_dots: number of point per each axis
    :return: tensor with function values
    :return: three tensor for drawing
    """
    _check_bounds(bounds)
    x_axis = torch.linspace(bounds[0][0], bounds[0][1], cnt_dots)
    y_axis = torch.linspace(bounds[1][0], bounds[1][1], cnt_dots)

    z_axis = _make_z(function, x_axis, y_axis, cnt_dots)

    return x_axis, y_axis, z_axis


def _make_descent_history(history: HistoryGD) -> pd.DataFrame:
    """
    Return converted HistoryGD object into pd.DataFrame with columsn ['x', 'y', 'z', 'iteration']::

        >>> point, hist = gd_optimal(lambda x: x[0] ** 2 + x[1] ** 2 * 1.01, torch.tensor([10, 10]), keep_history=True)
        >>> _make_descent_history(hist).round(4)
        +---+----------+----------+----------+-------------+
        |   | x        | y        | z        | iteration   |
        +===+==========+==========+==========+=============+
        | 0 | 10.0000  | 10.0000  | 201.000  | 0           |
        | 1 | 0.0502   | -0.0493  | 0.005    | 1           |
        | 2 | 0.0002   | 0.0002   | 0.000    | 2           |
        | 3 | 0.0000   | -0.0000  | 0.000    | 3           |
        +---+----------+----------+----------+-------------+

    :param history: History after some gradient method
    :return: pd.DataFrame

    """
    x_points, y_points = torch.stack(history['x']).T

    output_data = pd.DataFrame(
        {
            'x': x_points,
            'y': y_points,
            'z': torch.stack(history['f_value']),
            'iteration': history['iteration']
        }
    )
    return output_data


def _make_ranges(history: HistoryGD, k: float = 0.5) -> tuple[tuple[float, float], tuple[float, float]]:
    """
    Return bounds for the x-axis and the y-axis.
    1. Find a min_x and max_x and then x_range = [min_x - (max_x - min_x) * k, max_x + (max_x - min_x) * k]
    2. Similarly for y_axis

    :param history: History after some gradient method
    :param k: [min_x - (max_x - min_x) * k, max_x + (max_x - min_x) * k]
    :return: [x_range, y_range]
    """

    descent_history = _make_descent_history(history)
    min_x = descent_history.x.min() - 0.1
    max_x = descent_history.x.max() + 0.1

    min_y = descent_history.y.min() - 0.1
    max_y = descent_history.y.max() + 0.1

    x_range = (min_x - (max_x - min_x) * k, max_x + (max_x - min_x) * k)
    y_range = (min_y - (max_y - min_y) * k, max_y + (max_y - min_y) * k)

    return x_range, y_range


def _make_descent_frames_3d(function: Callable[[torch.Tensor], float],
                            history: HistoryGD,
                            number_points: int = 10) -> list[go.Frame]:
    """
    Make sequence of go.Frame which contain frame for each step of descent with a previous history

    :param function: callable that depends on the first positional argument
    :param history: History after some gradient method
    :param number_points: the number of points in will be calculated by a function from (x_i, y_i) to  (x_i+1, y_i+1)
    :return: list[go.Frame]
    """

    frames = []
    descent_history = _make_descent_history(history)

    draw_descent = [[], [], []]

    # For each step of optimization (minimization steps)
    for i in range(descent_history.shape[0]):

        if i > 0:
            # it will be calculated way from point(x_i, y_i) to point (x_i+1, y_i+1)
            x0, x1 = descent_history.x[i - 1], descent_history.x[i]
            y0, y1 = descent_history.y[i - 1], descent_history.y[i]

            if (x0 - x1) ** 2 + (y0 - y1) ** 2 > 0.5:
                num_p = number_points
            else:
                num_p = 2

            for alpha in torch.linspace(0, 1, num_p):
                draw_descent[0].append(x0 * alpha + x1 * (1 - alpha))
                draw_descent[1].append(y0 * alpha + y1 * (1 - alpha))
                draw_descent[2].append(function(torch.tensor([draw_descent[0][-1], draw_descent[1][-1]])))
            else:
                draw_descent[0].append(torch.nan)
                draw_descent[1].append(torch.nan)
                draw_descent[2].append(torch.nan)

        scatter_line = go.Scatter3d(x=draw_descent[0],
                                    y=draw_descent[1],
                                    z=draw_descent[2],
                                    name='descent',
                                    mode='lines',
                                    line={'width': 4, 'color': 'rgb(1, 23, 47)'})

        scatter_points = go.Scatter3d(x=descent_history.x[:i + 1],
                                      y=descent_history.y[:i + 1],
                                      z=descent_history.z[:i + 1],
                                      name='descent',
                                      mode='markers',
                                      marker={'size': 4, 'color': 'rgb(1, 23, 47)'},
                                      showlegend=False)

        frames.append(go.Frame(data=[scatter_points, scatter_line], name=i, traces=[1, 2]))

    return frames


def gen_animated_surface(function: Callable[[torch.Tensor], float],
                         history: HistoryGD,
                         cnt_dots: int = 100,
                         title: str = '<b>Surface with optimization steps</b>') -> go.Figure:
    """
    Return go.Figure with animation per each step of descent

    :param function: callable that depends on the first positional argument
    :param history: History after some gradient method
    :param cnt_dots: the numbers of point per each axis
    :param title: how many frames will drawing. ~300 frames will be drawn for ~5-10 seconds
    :return: go.Figure with animation steps on surface

    .. code-block:: python3

        >>> def f(x): return x[0] ** 2 + x[1] ** 2 / 9
        >>> _, h = bfgs(f, torch.tensor([10, 10]), keep_history=True)

    """
    descent_history = _make_descent_history(history)
    bounds = _make_ranges(history)

    first_point = go.Scatter3d(x=descent_history.x[:1],
                               y=descent_history.y[:1],
                               z=descent_history.z[:1],
                               mode='markers',
                               marker={'size': 4, 'color': 'rgb(1, 23, 47)'},
                               showlegend=False)

    surface = make_surface(function, bounds, cnt_dots=cnt_dots)
    layout = px.scatter_3d(descent_history, x='x', y='y', z='z', animation_frame='iteration').layout
    frames = _make_descent_frames_3d(function, history)

    fig = go.Figure(data=[surface, first_point, first_point],
                    layout=layout,
                    frames=frames)
    fig.update_scenes(
        xaxis_title=r'<b>x</b>',
        yaxis_title=r'<b>y</b>',
        zaxis_title=r'<b>z</b>',
    )
    camera = dict(
        up=dict(x=0, y=0, z=1),
        center=dict(x=0, y=0, z=0),
        eye=dict(x=-1.25, y=-1.25, z=1.25)
    )

    fig.update_layout(
        {
            'title': {
                'font': {'size': 22, 'color': 'black'},
                'text': title
            },
            'scene_camera': camera
        }
    )
    return fig



import sys
from typing import Optional, Callable

import numpy as np
import torch

if sys.version_info >= (3, 8):
    from typing import Literal, TypedDict
else:
    from typing_extensions import Literal, TypedDict


class NueSGD:

    def __init__(self, model: torch.nn.Module, lr: float = 1e-4):
        """
        Implementation of classic SGD (stochastic gradient descent) optimization algorithm.

        :param model: pytorch model that can be called and have a ".loss" method
        :param lr: learning rate. Multiplier of gradient step: x = x - lr * grad(x)
        """
        self.parameters = list(model.parameters())
        self.model = model
        self.lr = lr
        self.history = {'q_loss': []}

    @torch.no_grad()
    def step(self) -> None:
        """
        Update parameters data

        W = W - lr * Grad(W)

        :return: None
        """
        for param in self.parameters:
            param.data -= self.lr * param.grad.data

    @torch.no_grad()
    def zero_grad(self) -> None:
        """
        Make the gradients equal to zero

        :return: None
        """
        for param in self.parameters:
            if param.grad is not None:
                param.grad.data.zero_()

    def optimize(self,
                 x: torch.Tensor,
                 y: torch.Tensor,
                 epochs: int = 1,
                 batch_size: int = -1,
                 num_verbose: int = 0,
                 lamb: float = 0.3,
                 print_function: Callable = print) -> [torch.nn.Module, dict]:
        """
        Function apply MySGD optimizer, and train model.

        :param x: training set
        :param y: target value
        :param epochs: max number of sgd implements
        :param batch_size: size of batch for each epoch. default is -1 - all data
        :param num_verbose: number of iterations to be printed
        :param lamb: rate of history loss evaluation
        :param print_function: e.g. print or streamlit.write or something else
        :return: trained model and history
        """
        if batch_size == -1:
            batch_size = x.shape[0]

        q_new = self.model.loss(x, y)  # Q - functional evaluation
        print_epochs = np.geomspace(1, epochs + 1, num_verbose, dtype=int)

        for epoch in range(epochs):
            i = torch.randint(0, x.shape[0], [batch_size])  # choose batch

            # optimization
            self.zero_grad()
            loss = self.model.loss(x[i], y[i])
            loss.backward()
            self.step()

            # Q calculation
            q_pre = q_new
            q_new = q_pre * (1 - lamb) + loss.item() * lamb

            # history updating
            self.history['q_loss'].append(q_new.item())

            if epoch + 1 in print_epochs:
                print_function(f'epoch: {epoch + 1:5d} | Q: {q_new:0.4f}')

            if abs(q_new - q_pre) < 1e-6:
                break

        return self.model, self.history


class HistorySA(TypedDict):
    type_ball: tuple
    iteration: list
    point: Optional[list]
    loss: list


class SimulatedAnnealing:

    def __init__(self,
                 model: torch.nn.Module,
                 type_center: Literal['zero', 'neighborhood'] = 'neighborhood',
                 init_temp: float = 1_000_000,
                 radius: float = 1,
                 temp_multiplier: float = 0.95):
        """
        Initialization of SimulatedAnnealing algorithm. Minimize real number models (non-discrete)

        :param model: some pytorch model
        :param type_center: if type_center is zero, new point (x_k+1) would be chosen from Uniform[-radius, radius]
                            for each parameter, elif neighborhood, new point would be chosen from
                            Uniform[x_k - radius, x_k + radius).
        :param init_temp: initial temperature. Default is 10_000
        :param radius: ball's radius
        """

        self.temp = init_temp
        self.center = type_center
        self.radius = radius
        self.temp_multiplier = temp_multiplier
        self.history = {
            'type_ball': (type_center, radius),
            'iteration': [],
            'point': None if len(list(model.parameters())) > 1 else [],
            'loss': [],
            'best_point': None if len(list(model.parameters())) > 1 else [],
            'best_loss': []
        }
        self.best_state = model.state_dict()
        self.model = model
        self.min_temp = 1e-8
        self.best_loss = torch.inf
        self.init_temp = init_temp

    @torch.no_grad()
    def optimize_generator(self, x: torch.Tensor, y: torch.Tensor) -> str:
        """
        Generator of Simulated Annealing steps. [1]_

        :math:`\\rule{125mm}{0.7pt} \\\\`
        :math:`c = x_{pre} \\text{ if type area is `neighborhood' else } c = \\theta - \\text{zero} \\\\`
        :math:`x_{cur} \\sim \\mathcal{U}(c, r) \\qquad p \\sim \\mathcal{U}[0, 1]\\\\`

        :math:`\\text{if } f(x_{cur}) < f(x_{best}): \\\\`
        :math:`\\qquad x_{pre} = x_{best} = x_{cur}\\\\`

        :math:`\\text{elif } \\displaystyle \\exp\\left(\\frac{f(x_{pre}) - f(x_{cur})}{T}\\right) > p:\\\\`
        :math:`\\qquad x_{pre} = x_{cur}\\\\`

        :math:`T = T \\cdot \\delta`
        :math:`\\rule{125mm}{0.7pt} \\\\`

        :param x: training set
        :param y: target value
        :return: verbose strign with iteration and loss

        .. code-block:: python3

            >>> torch.random.manual_seed(7)

            >>> xr = torch.rand(100, 3)
            >>> w = torch.tensor([[1., 2., 3.]]).T
            >>> yr = xr @ w + 2

            >>> model = torch.nn.Sequential(torch.nn.Linear(3, 1))
            >>> model.loss = lambda _x, _y: torch.nn.MSELoss()(model(_x), _y)

            >>> optimizer = SimulatedAnnealing(model, temp_multiplier=0.01)

            >>> for verbose in optimizer.optimize(xr, yr):
            >>>     print(verbose)
            iteration:    1 | loss: 97.5745
            iteration:    2 | loss: 231.5806
            iteration:    3 | loss: 3.4633
            iteration:    4 | loss: 3.7009
            iteration:    5 | loss: 26.9238
            iteration:    6 | loss: 6.5509
            iteration:    7 | loss: 21.4261

            >>> model.loss(xr, yr)
            tensor(3.4633, grad_fn=<MseLossBackward0>)

        .. rubric:: References

        .. [1] Van Laarhoven, P. J. M., & Aarts, E. H. L. (1987). Simulated annealing: Theory and applications
               (1987th ed.). Kluwer Academic. pp.10-11

        """
        while self.temp > self.min_temp:

            pre_loss = self.model.loss(x, y).item()

            # init loss, iter, point
            if len(self.history['iteration']) == 0:
                self.best_loss = pre_loss
                self.history['iteration'].append(0)
                self.history['loss'].append(pre_loss)
                self.history['best_loss'].append(pre_loss)

                if self.history['point'] is not None:
                    for param in self.model.parameters():
                        self.history['point'].append(param.data)
                        self.history['best_point'].append(param.data)

            # choose new point
            for param in self.model.parameters():
                c = param.data if self.center == 'neighborhood' else torch.zeros_like(param.data)
                x_cur = (torch.rand_like(c) - 0.5) * 2 * self.radius + c
                param.data = x_cur

                if self.history['point'] is not None:
                    self.history['point'].append(param.data.flatten())

            # calc new loss
            cur_loss = self.model.loss(x, y).item()

            # check criterion
            if cur_loss < self.best_loss:
                self.best_state = self.model.state_dict()
                self.best_loss = cur_loss

            if cur_loss <= pre_loss:
                pass

            elif torch.e ** ((pre_loss - cur_loss) / self.temp) > torch.rand(1):
                pass

            else:
                self.model.load_state_dict(self.best_state)

            # update history
            self.history['iteration'].append(self.history['iteration'][-1] + 1)
            self.history['loss'].append(cur_loss)
            self.history['best_loss'].append(self.best_loss)

            if self.history['point'] is not None:
                self.history['best_point'].append(list(self.best_state.values())[0])

            # update temp
            self.temp *= self.temp_multiplier

            yield f'iteration: {self.history["iteration"][-1]:4d} | loss: {self.history["loss"][-1]:.4f}'

        else:
            # set best parameters
            self.model.load_state_dict(self.best_state)

    @torch.no_grad()
    def optimize(self, x: torch.Tensor, y: torch.Tensor) -> [torch.nn.Module, HistorySA]:

        for _ in self.optimize_generator(x, y):
            pass

        else:
            self.temp = self.init_temp

        return self.model, self.history



# Simple common math functions for show examples
# Gradient, Hessian, Jacobian

from __future__ import annotations

from typing import Callable, Sequence

import torch


def gradient(function: Callable[[torch.Tensor], torch.Tensor], x0: torch.Tensor, delta_x: float = 1e-4) -> torch.Tensor:
    """
    Returns the gradient of the function at a specific point :math:`x_0`.
    A two-point finite difference formula that approximates the derivative

    .. math::

        \\frac{\\partial f}{\\partial x_i} \\approx {\\frac {f(x_1, \\dots, x_i+h, \\dots, x_n)-
        f(x_1, \\dots, x_i-h, \\dots, x_n)}{2h}}

    Gradient

    .. math::

         \\displaystyle \\nabla f = \\left[\\frac{\\partial f}{\\partial x_1} \\enspace
         \\frac{\\partial f}{\\partial x_2}
         \\enspace \\dots \\enspace \\frac{\\partial f}{\\partial x_n}\\right]^\\top


    :param function: function which depends on n variables from x
    :param x0: n x 1 - dimensional array :math:`\\in \\mathbb{R}^{n}`. dtype is torch.double (float64)
    :param delta_x: precision of two-point formula above (delta_x = h)
    :return: vector of partial derivatives

    .. note::
        If we make delta_x :math:`\\leq` 1e-4 gradient will return values with large error rate

    .. rubric:: Examples

    .. code-block:: python3

        >>> # f(x, y)  = x ** 2 + y ** 2
        >>> gradient(lambda x: (x ** 2).sum(), torch.tensor([1., 2.]))
            tensor([2., 4.], dtype=torch.float64)

    """
    assert isinstance(x0, torch.Tensor), 'x0 must be torch.Tensor'

    x0: torch.Tensor = x0.flatten().double()
    delta_x = torch.tensor(max(delta_x, 1e-4), dtype=x0.dtype)
    grad = torch.zeros_like(x0, dtype=x0.dtype)

    for i in range(len(x0)):
        delta = torch.zeros_like(x0, dtype=x0.dtype)
        delta[i] += delta_x
        grad_i = (function(x0 + delta) - function(x0 - delta)) / (2 * delta_x)
        grad[i] = grad_i

    return grad


def jacobian(f_vector: Sequence[Callable[[torch.Tensor], torch.Tensor]],
             x0: torch.Tensor,
             delta_x: float = 1e-4) -> torch.Tensor:
    """
    Returns the Jacobian matrix of a sequence of m functions from f_vector by n variables from x.

    .. math::

        {\\displaystyle J ={\\begin{bmatrix}{\\dfrac {\\partial f_{1}}{\\partial x_{1}}}&\\cdots
          &{\\dfrac {\\partial f_{1}}{\\partial x_{n}}}\\\\\\vdots &\\ddots &\\vdots \\\\{\\dfrac {\\partial f_{m}}
          {\\partial x_{1}}}&\\cdots &{\\dfrac {\\partial f_{m}}{\\partial x_{n}}}\\end{bmatrix}}}_{m \\times n}


    :param f_vector: a flat sequence, list or tuple or other containing m functions
    :param x0: an n-dimensional array. The specific point at which we will calculate the Jacobian
    :param delta_x: precision of gradient
    :return: the Jacobian matrix according to the above formula. Matrix n x m

    .. rubric:: Examples

    .. code-block:: python3

        >>> func_3 = [lambda x: x[0] ** 2 + x[1], lambda x: 2 * x[0] + 5 * x[1], lambda x: x[0] * x[1]]
        >>> print(jacobian(func_3, torch.tensor([-1, 2])).round())
        tensor([[-2.,  1.],
                [ 2.,  5.],
                [ 2., -1.]], dtype=torch.float64)
    """
    assert isinstance(f_vector, Sequence), 'f_vector must be sequence'
    jac = torch.zeros(len(f_vector), x0.flatten().shape[0], dtype=torch.float64)
    for j in range(len(f_vector)):
        jac[j, :] = gradient(f_vector[j], x0, delta_x)

    return jac


def hessian(function: Callable[[torch.Tensor], torch.Tensor], x0: torch.Tensor, delta_x: float = 1e-4) -> torch.Tensor:
    """
    Returns a hessian of function at point :math:`x_0`


    .. math::
        \\ H(f) = \\begin{bmatrix} \\displaystyle
        \\frac{\\partial^2 f}{\\partial x_1^2} & \\displaystyle \\frac{\\partial^2 f}{\\partial x_1\\,\\partial x_2} &
        \\cdots & \\displaystyle \\frac{\\partial^2 f}{\\partial x_1\\,\\partial x_n} \\\\  \\
        \\displaystyle \\frac{\\partial^2 f}{\\partial x_2\\,\\partial x_1} & \\displaystyle \\frac{\\partial^2 f}
        {\\partial x_2^2} & \\cdots & \\displaystyle \\frac{\\partial^2 f}{\\partial x_2\\,\\partial x_n} \\\\  \\
        \\vdots & \\vdots & \\ddots & \\vdots \\\\  \\
        \\displaystyle \\frac{\\partial^2 f}{\\partial x_n\\,\\partial x_1} & \\displaystyle \\frac{\\partial^2 f}
        {\\partial x_n\\,\\partial x_2} & \\cdots & \\displaystyle \\frac{\\partial^2 f}{\\partial x_n^2}
        \\end{bmatrix}\\\\




    :param function: function which depends on n variables from x
    :param x0: n - dimensional array
    :param delta_x: precision of two-point formula above (delta_x = h)
    :return: the hessian of function

    .. note::
        If we make delta_x :math:`\\leq` 1e-4 hessian returns matrix with large error rate



    .. rubric:: Examples

    .. code-block:: python3

        >>> def paraboloid(x): return x[0] ** 2 + 2 * x[1] ** 2
        >>> print(hessian(paraboloid, torch.tensor([1, 1])).round())
        [[2. 0.]
         [0. 4.]]
    """
    delta_x = max(delta_x, 1e-4)  # Check note
    x0: torch.Tensor = x0.flatten().double()
    hes = torch.zeros(x0.shape[0], x0.shape[0], dtype=torch.float64)

    for i in range(len(x0)):
        delta_i = torch.zeros_like(x0, dtype=torch.float64)
        delta_i[i] += delta_x

        def partial_i(x: torch.Tensor) -> torch.Tensor:
            return (function(x + delta_i) - function(x - delta_i)) / (2 * delta_x)

        hes[i, :] = gradient(partial_i, x0, delta_x)

    return hes
